from functools import partial
from typing import Optional

import click
import numpy as np
from trackpy import bandpass

from starfish.imagestack.imagestack import ImageStack
from starfish.types import Number
from ._base import FilterAlgorithmBase
from .util import determine_axes_to_group_by, preserve_float_range


class Bandpass(FilterAlgorithmBase):

    def __init__(
            self, lshort: Number, llong: int, threshold: Number=0, truncate: Number=4,
            is_volume: bool=False) -> None:
        """

        Parameters
        ----------
        lshort : float
            filter frequencies below this value
        llong : int
            filter frequencies above this odd integer value
        threshold : float
            zero any spots below this intensity value after background subtraction (default 0)
        truncate : float
            truncate the gaussian kernel, used by the gaussian filter, at this many standard
            deviations (default 4)
        is_volume : bool
            If True, 3d (z, y, x) volumes will be filtered. By default, filter 2-d (y, x) planes
        """
        self.lshort = lshort
        self.llong = llong

        if threshold is None:
            raise ValueError("Threshold cannot be None, please pass a float or integer")

        self.threshold = threshold
        self.truncate = truncate
        self.is_volume = is_volume

    _DEFAULT_TESTING_PARAMETERS = {"lshort": 1, "llong": 3, "threshold": 0.01}

    @staticmethod
    def _bandpass(
            image: np.ndarray, lshort: Number, llong: int, threshold: Number, truncate: Number
    ) -> np.ndarray:
        """Apply a bandpass filter to remove noise and background variation

        Parameters
        ----------
        image : np.ndarray
        lshort : float
            filter frequencies below this value
        llong : int
            filter frequencies above this odd integer value
        threshold : float
            zero any spots below this intensity value after background subtraction (default 0)
        truncate : float
            truncate the gaussian kernel, used by the gaussian filter, at this many standard
            deviations

        Returns
        -------
        np.ndarray :
            bandpassed image

        """
        bandpassed: np.ndarray = bandpass(
            image, lshort=lshort, llong=llong, threshold=threshold,
            truncate=truncate
        )
        return preserve_float_range(bandpassed)

    def run(
            self, stack: ImageStack, in_place: bool = False, verbose: bool = False,
            n_processes: Optional[int] = None
    ) -> ImageStack:
        """Perform filtering of an image stack

        Parameters
        ----------
        stack : ImageStack
            Stack to be filtered.
        in_place : bool
            if True, process ImageStack in-place, otherwise return a new stack
        verbose : bool
            if True, report the filtering progress across the tiles or volumes of the ImageStack
        n_processes : Optional[int]
            Number of parallel processes to devote to calculating the filter

        Returns
        -------
        ImageStack :
            If in-place is False, return the results of filter as a new stack.  Otherwise return the
            original stack.

        """
        bandpass_ = partial(
            self._bandpass,
            lshort=self.lshort, llong=self.llong, threshold=self.threshold, truncate=self.truncate
        )

        group_by = determine_axes_to_group_by(self.is_volume)

        result = stack.apply(
            bandpass_,
            group_by=group_by,
            in_place=in_place,
            n_processes=n_processes,
        )
        return result

    @staticmethod
    @click.command("Bandpass")
    @click.option(
        "--lshort", type=float, help="filter signals below this frequency")
    @click.option(
        "--llong", type=int, help="filter signals above this frequency")
    @click.option(
        "--threshold", default=0, type=float, help="zero pixels below this intensity value")
    @click.option(
        "--truncate", default=4, type=float,
        help="truncate the filter at this many standard deviations")
    @click.pass_context
    def _cli(ctx, lshort, llong, threshold, truncate):
        ctx.obj["component"]._cli_run(ctx, Bandpass(lshort, llong, threshold, truncate))



from copy import deepcopy
from typing import Optional

import click
import numpy as np
from tqdm import tqdm

from starfish.imagestack.imagestack import ImageStack
from starfish.types import Indices
from ._base import FilterAlgorithmBase


class ZeroByChannelMagnitude(FilterAlgorithmBase):
    def __init__(self, thresh: int, normalize: bool) -> None:
        """For assays in which we expect codewords to have explicit zero values,
        e.g., DARTFISH, SEQFISH, etc., this filter allows for the explicit zeroing
        out of pixels, for each round, where there is insufficient signal magnitude across channels.


        Parameters
        ----------
        thresh : int
            pixels in each round that have a L2 norm across channels below this threshold
            are set to 0

        normalize : bool
            if True, this scales all rounds to have unit L2 norm across channels
        """
        self.thresh = thresh
        self.normalize = normalize

    _DEFAULT_TESTING_PARAMETERS = {"thresh": 0, "normalize": True}

    def run(
            self, stack: ImageStack,
            in_place: bool=False, verbose=False, n_processes: Optional[int]=None
    ) -> ImageStack:
        """Perform filtering of an image stack

        Parameters
        ----------
        stack : ImageStack
            Stack to be filtered.
        in_place : bool
            if True, process ImageStack in-place, otherwise return a new stack
        verbose : bool
            if True, report on the percentage completed during processing (default = False)
        n_processes : Optional[int]: None
            Not implemented. Number of processes to use when applying filter.

        Returns
        -------
        ImageStack :
            If in-place is False, return the results of filter as a new stack.  Otherwise return the
            original stack.

        """
        channels_per_round = stack._data.groupby(Indices.ROUND.value)
        channels_per_round = tqdm(channels_per_round) if verbose else channels_per_round

        if not in_place:
            new_stack = deepcopy(stack)
            return self.run(new_stack, in_place=True)

        # compute channel magnitude mask
        for r, dat in channels_per_round:
            # nervous about how xarray orders dimensions so i put this here explicitly ....
            dat = dat.transpose(Indices.CH.value,
                                Indices.Z.value,
                                Indices.Y.value,
                                Indices.X.value
                                )
            # ... to account for this line taking the norm across axis 0, or the channel axis
            ch_magnitude = np.linalg.norm(dat, ord=2, axis=0)
            magnitude_mask = ch_magnitude >= self.thresh

            # apply mask and optionally, normalize by channel magnitude
            for c in range(stack.num_chs):
                ind = {Indices.ROUND.value: r, Indices.CH.value: c}
                stack._data[ind] = stack._data[ind] * magnitude_mask

                if self.normalize:
                    stack._data[ind] = np.divide(stack._data[ind],
                                                 ch_magnitude,
                                                 where=magnitude_mask
                                                 )
        return stack

    @staticmethod
    @click.command("ZeroByChannelMagnitude")
    @click.option(
        '--thresh', type=float,
        help='minimum magnitude threshold for pixels across channels')
    @click.option(
        '--normalize', is_flag=True,
        help='Scales all rounds to have unit L2 norm across channels')
    @click.pass_context
    def _cli(ctx, thresh, normalize):
        ctx.obj["component"]._cli_run(ctx, ZeroByChannelMagnitude(thresh, normalize))



import unittest
from collections import OrderedDict

import numpy as np
import xarray as xr

from starfish.imagestack import physical_coordinate_calculator
from starfish.types import Indices, PHYSICAL_COORDINATE_DIMENSION, PhysicalCoordinateTypes


X_INDEX_0 = 1
X_INDEX_1 = (.005 * 1) + 1
X_INDEX_100 = (.005 * 100) + 1
X_INDEX_101 = (.005 * 101) + 1
X_INDEX_151 = (.005 * 151) + 1
Y_INDEX_0 = 4
Y_INDEX_100 = (.01 * 100) + 4
Y_INDEX_101 = (.01 * 101) + 4
Y_INDEX_191 = (.01 * 191) + 4
Y_INDEX_200 = 6


class TestPhysicalCoordinateCalculator(unittest.TestCase):
    """Tests recalculating physical coordinates for an Imagestack with a shape (1, 1, 1, 200, 200)
    and physical coordinates {xmin: 1 , xmax: 2, ymin: 4, ymax: 6, zmin: 1, zmax: 3}"""

    def setUp(self):
        self.coords_array = xr.DataArray(
            np.empty(
                shape=(1, 1, 1, 6),
                dtype=np.float32,
            ),
            dims=(Indices.ROUND.value,
                  Indices.CH.value,
                  Indices.Z.value,
                  PHYSICAL_COORDINATE_DIMENSION),
            coords={
                PHYSICAL_COORDINATE_DIMENSION: [
                    PhysicalCoordinateTypes.X_MIN.value,
                    PhysicalCoordinateTypes.X_MAX.value,
                    PhysicalCoordinateTypes.Y_MIN.value,
                    PhysicalCoordinateTypes.Y_MAX.value,
                    PhysicalCoordinateTypes.Z_MIN.value,
                    PhysicalCoordinateTypes.Z_MAX.value,
                ],
            },
        )

        self.indexers = {Indices.ROUND: slice(None, None),
                         Indices.CH: slice(None, None),
                         Indices.Z: slice(None, None),
                         Indices.Y: slice(None, None),
                         Indices.X: slice(None, None)}

        self.physical_coords = {PhysicalCoordinateTypes.X_MIN: 1.0,
                                PhysicalCoordinateTypes.X_MAX: 2.0,
                                PhysicalCoordinateTypes.Y_MIN: 4.0,
                                PhysicalCoordinateTypes.Y_MAX: 6.0,
                                PhysicalCoordinateTypes.Z_MIN: 1.0,
                                PhysicalCoordinateTypes.Z_MAX: 3.0}

        self.coords_array.loc[0, 0, 0] = \
            np.array([self.physical_coords[PhysicalCoordinateTypes.X_MIN],
                      self.physical_coords[PhysicalCoordinateTypes.X_MAX],
                      self.physical_coords[PhysicalCoordinateTypes.Y_MIN],
                      self.physical_coords[PhysicalCoordinateTypes.Y_MAX],
                      self.physical_coords[PhysicalCoordinateTypes.Z_MIN],
                      self.physical_coords[PhysicalCoordinateTypes.Z_MAX]])

        self.stack_shape = OrderedDict([(Indices.ROUND, 1), (Indices.CH, 1),
                                        (Indices.Z, 1), (Indices.Y, 200), (Indices.X, 200)])

    def test_calc_new_physical_coords_array_single_y_slice_x(self):
        # Index on single value of X, range of y
        self.indexers[Indices.Y], self.indexers[Indices.X] = 100, slice(None, 100)
        new_coords = physical_coordinate_calculator.calc_new_physical_coords_array(
            physical_coordinates=self.coords_array,
            stack_shape=self.stack_shape,
            indexers=self.indexers)

        expected_xmin, expected_xmax = X_INDEX_0, X_INDEX_101
        expected_ymin, expected_ymax = Y_INDEX_100, Y_INDEX_101

        # z range stays the same
        expected_zmin = self.physical_coords[PhysicalCoordinateTypes.Z_MIN]
        expected_zmax = self.physical_coords[PhysicalCoordinateTypes.Z_MAX]

        assert np.allclose(new_coords.loc[0, 0, 0], np.array([expected_xmin,
                                                              expected_xmax,
                                                              expected_ymin,
                                                              expected_ymax,
                                                              expected_zmin,
                                                              expected_zmax]))

    def test_calc_new_physical_coords_array_slice_x_y(self):
        # Index on last half of Y, first half of X
        self.indexers[Indices.Y], self.indexers[Indices.X] = slice(100, None), slice(None, 100)
        new_coords = physical_coordinate_calculator.calc_new_physical_coords_array(
            physical_coordinates=self.coords_array,
            stack_shape=self.stack_shape,
            indexers=self.indexers)

        expected_xmin, expected_xmax = X_INDEX_0, X_INDEX_101
        expected_ymin, expected_ymax = Y_INDEX_100, Y_INDEX_200

        # z range stays the same
        expected_zmin = self.physical_coords[PhysicalCoordinateTypes.Z_MIN]
        expected_zmax = self.physical_coords[PhysicalCoordinateTypes.Z_MAX]

        assert np.allclose(new_coords.loc[0, 0, 0], np.array([expected_xmin,
                                                              expected_xmax,
                                                              expected_ymin,
                                                              expected_ymax,
                                                              expected_zmin,
                                                              expected_zmax]))

    def test_calc_new_physical_coords_array_negative_indexing(self):
        # Negative indexing
        self.indexers[Indices.Y], self.indexers[Indices.X] = slice(100, -10), slice(1, -50)
        new_coords = physical_coordinate_calculator.calc_new_physical_coords_array(
            physical_coordinates=self.coords_array,
            stack_shape=self.stack_shape,
            indexers=self.indexers)

        expected_xmin, expected_xmax = X_INDEX_1, X_INDEX_151
        expected_ymin, expected_ymax = Y_INDEX_100, Y_INDEX_191

        # z range stays the same
        expected_zmin = self.physical_coords[PhysicalCoordinateTypes.Z_MIN]
        expected_zmax = self.physical_coords[PhysicalCoordinateTypes.Z_MAX]

        assert np.allclose(new_coords.loc[0, 0, 0], np.array([expected_xmin,
                                                              expected_xmax,
                                                              expected_ymin,
                                                              expected_ymax,
                                                              expected_zmin,
                                                              expected_zmax]))


if __name__ == '__main__':
    unittest.main()



from typing import Type

import click

from starfish.codebook.codebook import Codebook
from starfish.intensity_table.intensity_table import IntensityTable
from starfish.pipeline import AlgorithmBase, PipelineComponent
from . import _base
from . import per_round_max_channel_decoder


class Decoder(PipelineComponent):

    @classmethod
    def _get_algorithm_base_class(cls) -> Type[AlgorithmBase]:
        return _base.DecoderAlgorithmBase

    @classmethod
    def _cli_run(cls, ctx, instance):
        table = ctx.obj["intensities"]
        codes = ctx.obj["codebook"]
        output = ctx.obj["output"]
        intensities = instance.run(table, codes)
        intensities.save(output)

    @staticmethod
    @click.group("decode")
    @click.option("-i", "--input", required=True, type=click.Path(exists=True))
    @click.option("-o", "--output", required=True)
    @click.option("--codebook", required=True, type=click.Path(exists=True))
    @click.pass_context
    def _cli(ctx, input, output, codebook):
        ctx.obj = dict(
            component=Decoder,
            input=input,
            output=output,
            intensities=IntensityTable.load(input),
            codebook=Codebook.from_json(codebook),
        )


Decoder._cli_register()



import os
import subprocess
import tempfile
from typing import Callable, Sequence, Union

from starfish.util import clock


def stages(commands: Sequence[Sequence[Union[str, Callable]]],
           subdirs: Sequence[str]=None,

           keep_data: bool=False) -> str:
    """
    Execute a list of commands in a temporary directory
    cleaning them up unless otherwise requested.

    Parameters
    ----------
    commands : Sequence[Sequence[Union[str, Callable]]]
        A collection of tuples of commands composed either of
        str elements or callable objects which will be invoked
        with the keyword argument "tempdir". The resulting list
        of strings will be passed to subprocess.check_call.
    subdirs : Sequence[str]
        A collection of paths which should be created as subdirectories
        within the temporary directory used by this invocation.
    keep_data : bool
        If not true, shutil.rmtree will be called on the temporary
        directory used by this invocation.
    STARFISH_COVERAGE :
        This parameter is read from the environment. If set, then command lists will have
        `coverage run ...` options prepended before execution.

    Return
    ------
    str :
        Path to the temporary directory used by this invocation.

    """

    if keep_data:
        tempobj = None
        tempdir = tempfile.mkdtemp()
    else:
        tempobj = tempfile.TemporaryDirectory()
        tempdir = tempobj.name

    def callback(interval):
        print(" ".join(stage[:2]), " ==> {} seconds".format(interval))

    try:

        if subdirs:
            for subdir in subdirs:
                os.makedirs("{tempdir}".format(
                    tempdir=os.path.join(tempdir, subdir)))

        for stage in commands:
            cmdline = prepare_stage(stage, tempdir)
            with clock.timeit(callback):
                subprocess.check_call(cmdline)

        return tempdir

    finally:
        if tempobj:
            tempobj.cleanup()


def prepare_stage(stage: Sequence[Union[str, Callable]],
                  tempdir: str) -> Sequence[str]:
    """
    Loop through elements of stage, building them into a commandline.
    If an element is a callable, it will be invoked with the "tempdir"
    keyword.

    Parameters
    ----------
    stage: Sequence[Union[str, Callable]]
        A collection of commands composed either of
        str elements or callable objects which will be invoked
        with the keyword argument "tempdir". The resulting list
        of strings will be passed to subprocess.check_call.
    tempdir: str
        Temporary directory that will be used by the invoking method.
    STARFISH_COVERAGE :
        This parameter is read from the environment. If set, then command lists will have
        `coverage run ...` options prepended before execution.

    Return
    ------
    Sequence of strings for passing to subprocess.check_call

    """
    coverage_enabled = "STARFISH_COVERAGE" in os.environ
    cmdline = [
        element(tempdir=tempdir) if callable(element) else element
        for element in stage
    ]
    if cmdline[0] == "starfish" and coverage_enabled:
        coverage_cmdline = [
            "coverage", "run",
            "-p",
            "--source", "starfish",
            "-m", "starfish",
        ]
        coverage_cmdline.extend(cmdline[1:])
        cmdline = coverage_cmdline
    return cmdline



import numpy as np

from starfish import ImageStack, IntensityTable
from starfish.intensity_table.concatenate import concatenate


def test_intensity_table_concatenation():
    """create two IntensityTables and assert that they are being concatenated properly."""

    r, c, z, y, x = 3, 3, 2, 2, 5
    data = np.zeros(180, dtype=np.float32).reshape(r, c, z, y, x)
    image_stack = ImageStack.from_numpy_array(data)
    intensities = IntensityTable.from_image_stack(image_stack)

    intensities2 = intensities.copy()

    original_shape = intensities.shape

    expected_shape = list(original_shape)
    expected_shape[0] *= 2  # only features is concatenated
    assert np.array_equal(
        concatenate([intensities, intensities2]).shape,
        expected_shape
    )

    # slice out a single channel and round from both experiments, such that the data no longer match
    # across all dimensions but the concatenation dimension. The resulting structure should be
    # 2 (r) * 2 (c) * 5 (z), 2, 2 = 40, 2, 2
    i1 = intensities.where(np.logical_and(intensities.r == 0, intensities.c == 0), drop=True)
    i2 = intensities.where(np.logical_and(intensities.r == 1, intensities.c == 1), drop=True)
    expected_shape = (i1.shape[0] + i2.shape[0], 2, 2)
    result = concatenate([i1, i2])

    assert expected_shape == result.shape

    # slice a larger r value for second array, however, there are still only two values, so
    # shape should be 40, 2, 2
    i3 = intensities.where(np.logical_and(intensities.r == 2, intensities.c == 1), drop=True)
    expected_shape = (i1.shape[0] + i3.shape[0], 2, 2)
    result = concatenate([i1, i3])

    assert expected_shape == result.shape

    # slice out z in addition to reduce the total feature number by 1/2
    i4 = intensities.where(np.logical_and(intensities.r == 0, intensities.z == 1), drop=True)
    expected_shape = (i1.shape[0] + i4.shape[0], 3, 1)
    result = concatenate([i1, i4])

    assert expected_shape == result.shape



from typing import Iterable

import xarray as xr

from starfish.types import Features
from .intensity_table import IntensityTable


def concatenate(intensity_tables: Iterable[IntensityTable]) -> IntensityTable:
    """Concatenate IntensityTables produced for different fields of view or across imaging rounds

    IntensityTables are concatenated along the features axis, and the resulting arrays are stored
    densely, even if the underlying data is sparse, since xarray does not yet support sparse array
    structures. This means that spots that are identified in different rounds and channels will
    be identified as separate features, even if they have exactly identical coordinates.

    To merge spots that share coordinates across rounds and channels into single features amenable
    to decoding, use IntensityTable.combine_first()

    Parameters
    ----------
    intensity_tables : Iterable[IntensityTable]
        iterable (list-like) of intensity tables to combine

    Returns
    -------
        merged IntensityTable. Missing values are filled with np.NaN

    See Also
    --------
    Sparse Arrays in xarray: https://github.com/pydata/xarray/issues/1375
    Combine_first: http://xarray.pydata.org/en/stable/combining.html#combine
    """
    return xr.concat(list(intensity_tables), Features.AXIS)



import xarray as xr

from starfish.util.try_import import try_import


class ExpressionMatrix(xr.DataArray):

    """Container for expression data extracted from an IntensityTable

    An ExpressionMatrix is a 2-dimensional ``cells`` x ``genes`` tensor whose values are the
    number of spots observed for each gene observed by the experiment. In addition to the basic
    xarray methods, IntensityTable implements:

    Methods
    -------
    save(filename)
        save the ExpressionMatrix to netCDF

    save_loom(filename)
        save the ExpressionMatrix to loom for use in R or python

    save_anndata(filename)
        save the ExpressionMatrix to AnnData for use in ``Scanpy``

    load(filename)
        load an ExpressionMatrix from netCDF
    """

    def save(self, filename: str) -> None:
        """Save an ExpressionMatrix as a Netcdf File

        Parameters
        ----------
        filename : str
            Name of Netcdf file
        """
        self.to_netcdf(filename)

    @try_import({"loompy"})
    def save_loom(self, filename: str) -> None:
        """Save an ExpressionMatrix as a loom file

        Parameters
        ----------
        filename : str
            Name of loom file
        """
        import loompy

        row_attrs = {k: self['cells'][k].values for k in self['cells'].coords}
        col_attrs = {k: self['genes'][k].values for k in self['genes'].coords}

        loompy.create(filename, self.data, row_attrs, col_attrs)

    @try_import({"anndata"})
    def save_anndata(self, filename: str) -> None:
        """Save an ExpressionMatrix as an AnnData file

        Parameters
        ----------
        filename : str
            Name of AnnData file
        """
        import anndata

        row_attrs = {k: self['cells'][k].values for k in self['cells'].coords}
        col_attrs = {k: self['genes'][k].values for k in self['genes'].coords}
        anndata = anndata.AnnData(self.data, row_attrs, col_attrs)
        anndata.write(filename)

    @classmethod
    def load(cls, filename: str) -> "ExpressionMatrix":
        """load an ExpressionMatrix from Netcdf

        Parameters
        ----------
        filename : str
            File to load

        Returns
        -------
        ExpressionMatrix

        """
        loaded = xr.open_dataarray(filename)
        expression_matrix = cls(
            loaded.data,
            loaded.coords,
            loaded.dims
        )
        return expression_matrix



from functools import partial
from typing import Optional

import click
import numpy as np
from scipy.signal import convolve, fftconvolve

from starfish.imagestack.imagestack import ImageStack
from starfish.types import Indices, Number
from ._base import FilterAlgorithmBase
from .util import gaussian_kernel, preserve_float_range


class DeconvolvePSF(FilterAlgorithmBase):

    def __init__(
            self, num_iter: int, sigma: Number, clip: bool=True) -> None:
        """Deconvolve a point spread function

        Parameters
        ----------
        num_iter : int
            number of iterations to run. Note that this is a very important parameter that requires
            careful optimization
        sigma : Number
            standard deviation of the gaussian kernel used to construct the point spread function
        clip : bool (default = False)
            if True, pixel values below -1 and above 1 are clipped for skimage pipeline
            compatibility

        """
        self.num_iter = num_iter
        self.sigma = sigma
        self.clip = clip
        self.kernel_size: int = int(2 * np.ceil(2 * sigma) + 1)
        self.psf: np.ndarray = gaussian_kernel(
            shape=(self.kernel_size, self.kernel_size),
            sigma=sigma
        )

    _DEFAULT_TESTING_PARAMETERS = {"num_iter": 1, "sigma": 1}

    # Here be dragons. This algorithm had a bug, but the results looked nice. Now we've "fixed" it
    # and the results look bad. #548 addresses this problem.
    @staticmethod
    def _richardson_lucy_deconv(
            image: np.ndarray, iterations: int, psf: np.ndarray, clip: bool) -> np.ndarray:
        """
        Deconvolves input image with a specified point spread function.

        Parameters
        ----------
        image : ndarray
           Input degraded image (can be N dimensional).
        psf : ndarray
           The point spread function.
        iterations : int
           Number of iterations. This parameter plays the role of
           regularisation.
        clip : boolean
            If true, pixel value of the result above 1 or
           under -1 are thresholded for skimage pipeline compatibility.

        Returns
        -------
        im_deconv : ndarray
           The deconvolved image.

        Examples
        --------
        >>> from skimage import color, data, restoration
        >>> camera = color.rgb2gray(data.camera())
        >>> from scipy.signal import convolve2d
        >>> psf = np.ones((5, 5)) / 25
        >>> camera = convolve2d(camera, psf, 'same')
        >>> camera += 0.1 * camera.std() * np.random.standard_normal(camera.shape)
        >>> deconvolved = restoration.richardson_lucy(camera, psf, 5)

        References
        ----------
        .. [1] http://en.wikipedia.org/wiki/Richardson%E2%80%93Lucy_deconvolution

        Notes
        -----
        This code is copied from skimage.restoration. We copied it to implement a bugfix wherein
        zeros in the input image or zeros produced during an intermediate would induce divide by
        zero -> Nan. These Nans would then propagate throughout the image, invalidating the results.
        Longer term, we will make a PR to skimage to introduce the fix. There is some existing work
        linked here: https://github.com/scikit-image/scikit-image/issues/2551

        """
        # compute the times for direct convolution and the fft method. The fft is of
        # complexity O(N log(N)) for each dimension and the direct method does
        # straight arithmetic (and is O(n*k) to add n elements k times)
        direct_time = np.prod(image.shape + psf.shape)
        fft_time = np.sum([n * np.log(n) for n in image.shape + psf.shape])

        # see whether the fourier transform convolution method or the direct
        # convolution method is faster (discussed in scikit-image PR #1792)
        time_ratio = 40.032 * fft_time / direct_time

        if time_ratio <= 1 or len(image.shape) > 2:
            convolve_method = fftconvolve
        else:
            convolve_method = convolve

        image = image.astype(np.float)
        psf = psf.astype(np.float)
        im_deconv = 0.5 * np.ones(image.shape)
        psf_mirror = psf[::-1, ::-1]

        eps = np.finfo(image.dtype).eps
        for _ in range(iterations):
            x = convolve_method(im_deconv, psf, 'same')
            np.place(x, x == 0, eps)
            relative_blur = image / x + eps
            im_deconv *= convolve_method(relative_blur, psf_mirror, 'same')

        if np.all(np.isnan(im_deconv)):
            raise RuntimeError(
                'All-NaN output data detected. Likely cause is that deconvolution has been run for '
                'too many iterations.')

        if clip:
            # Changing to cliping values above 1 here changes test results
            # so keeping this as rescaling for now
            im_deconv = preserve_float_range(im_deconv, True)

        return im_deconv

    def run(
            self, stack: ImageStack, in_place: bool=False, verbose=False,
            n_processes: Optional[int]=None
    ) -> ImageStack:
        """Perform filtering of an image stack

        Parameters
        ----------
        stack : ImageStack
            Stack to be filtered.
        in_place : bool
            if True, process ImageStack in-place, otherwise return a new stack
        verbose : bool
            if True, report on the percentage completed during processing (default = False)
        n_processes : Optional[int]
            Number of parallel processes to devote to calculating the filter

        Returns
        -------
        ImageStack :
            If in-place is False, return the results of filter as a new stack.  Otherwise return the
            original stack.

        """
        func = partial(
            self._richardson_lucy_deconv,
            iterations=self.num_iter, psf=self.psf, clip=self.clip
        )
        result = stack.apply(
            func,
            group_by={Indices.ROUND, Indices.CH, Indices.Z},
            verbose=verbose,
            n_processes=n_processes,
            in_place=in_place,
        )
        return result

    @staticmethod
    @click.command("DeconvolvePSF")
    @click.option(
        '--num-iter', type=int, help='number of iterations to run')
    @click.option(
        '--sigma', type=float, help='standard deviation of gaussian kernel')
    @click.option(
        '--no-clip', is_flag=True,
        help='(default True) if True, clip values below 0 and above 1')
    @click.pass_context
    def _cli(ctx, num_iter, sigma, no_clip):
        ctx.obj["component"]._cli_run(ctx, DeconvolvePSF(num_iter, sigma, no_clip))



from json import loads

from pytest import raises

from starfish.util.config import Config


simple_str = '{"a": 1}'
simple_map = loads(simple_str)

deep_str = '{"a": {"b": {"c": [1, 2, 3]}}}'
deep_map = loads(deep_str)

def test_simple_config_value_str():
    config = Config(simple_str)
    assert config.data["a"] == 1


def test_simple_config_value_map():
    config = Config(simple_map)
    assert config.data["a"] == 1


def test_simple_config_value_default_key(monkeypatch):
    monkeypatch.setenv("STARFISH_CONFIG", simple_str)
    config = Config()
    assert config.data["a"] == 1


def test_simple_config_value_file(tmpdir):
    f = tmpdir.join("config.json")
    f.write(simple_str)
    config = Config(f"@{f}")
    assert config.data["a"] == 1


def test_lookup_dne():
    config = Config(simple_str)
    with raises(KeyError):
        config.lookup(["foo"])
    assert config.lookup(["foo"], 1) == 1
    assert config.lookup(["foo", "bar"], 2) == 2


def test_lookup_deep():
    config = Config(deep_str)
    assert config.lookup(["a"]) == {"b": {"c": [1, 2, 3]}}
    assert config.lookup(["a", "b"]) == {"c": [1, 2, 3]}
    assert config.lookup(["a", "b", "c"]) == [1, 2, 3]
    with raises(AttributeError):
        config.lookup(["a", "b", "c", "d"])
    assert config.lookup(["a", "b", "c", "d"], "x") == "x"



"""
Tests for codebook.metric_decode method
"""

import numpy as np
import pandas as pd
import pytest

from starfish import Codebook, IntensityTable
from starfish.types import Features, Indices, SpotAttributes


def intensity_table_factory(data: np.ndarray=np.array([[[0, 3], [4, 0]]])) -> IntensityTable:
    """
    Produces an IntensityTable with a single feature that was measured over 2 channels and 2 rounds.
    """

    # generates spot attributes equal in size to the number of passed features.
    # each attribute has coordinates (z, y, x) equal to the feature index, and radius 1.
    spot_attributes_data = pd.DataFrame(
        data=np.array([[i, i, i, 1] for i in np.arange(data.shape[0])]),
        columns=[Indices.Z, Indices.Y, Indices.X, Features.SPOT_RADIUS]
    )

    intensity_table = IntensityTable.from_spot_data(data, SpotAttributes(spot_attributes_data))
    return intensity_table


def codebook_factory() -> Codebook:
    """
    Codebook with two codewords describing an experiment with three channels and two imaging rounds.
    Both codes have two "on" channels.
    """
    codebook_array = [
        {
            Features.CODEWORD: [
                {Indices.ROUND.value: 0, Indices.CH.value: 1, Features.CODE_VALUE: 1},
                {Indices.ROUND.value: 1, Indices.CH.value: 0, Features.CODE_VALUE: 1}
            ],
            Features.TARGET: 'GENE_A'
        },
        {
            Features.CODEWORD: [
                {Indices.ROUND.value: 1, Indices.CH.value: 0, Features.CODE_VALUE: 1},
                {Indices.ROUND.value: 1, Indices.CH.value: 1, Features.CODE_VALUE: 1}
            ],
            Features.TARGET: 'GENE_B'
        },
    ]
    return Codebook.from_code_array(codebook_array)


def test_metric_decode():
    """
    This test exposes 3 test features, each the the same normalized trace.
    The first should decode to GENE_A, and pass both the intensity and distance filters
    The second should decode to GENE_B, but fail the intensity filter
    The third should decode to GENE_B, as it is less far from that gene than GENE_A, but
    should nevertheless fail the distance filter because the tiles other than (0, 0) don't
    match
    """
    data = np.array(
        [[[0, 3],  # this code is decoded "right"
          [4, 0]],
         [[0, 0.4],  # this code should be filtered based on magnitude
          [0, 0.3]],
         [[30, 0],  # this code should be filtered based on distance
          [0, 40]]]
    )
    intensities = intensity_table_factory(data)
    codebook = codebook_factory()

    decoded_intensities = codebook.metric_decode(
        intensities,
        max_distance=0.5,
        min_intensity=1,
        norm_order=1
    )

    assert hasattr(decoded_intensities, Features.DISTANCE)

    assert decoded_intensities.sizes[Features.AXIS] == 3

    assert np.array_equal(
        decoded_intensities[Features.TARGET].values,
        ['GENE_A', 'GENE_B', 'GENE_B'],
    )

    assert np.array_equal(
        decoded_intensities[Features.PASSES_THRESHOLDS].values,
        [True, False, False]
    )

    assert not np.all(decoded_intensities == intensities)


def test_unmatched_intensities_and_codebook_table_sizes_throws_value_error():
    """
    Codebook and Intensity channel and round number must match. Here we use a codebook with 3
    channels, but an IntensityTable with only 2 to verify an error is thrown.
    """

    # this codebook has 3 channels
    codebook_array = [
        {
            Features.CODEWORD: [
                {Indices.ROUND.value: 0, Indices.CH.value: 2, Features.CODE_VALUE: 1},
                {Indices.ROUND.value: 1, Indices.CH.value: 0, Features.CODE_VALUE: 1}
            ],
            Features.TARGET: 'GENE_A'
        },
    ]
    codebook = Codebook.from_code_array(codebook_array)
    intensities = intensity_table_factory()
    with pytest.raises(ValueError):
        codebook.metric_decode(intensities, max_distance=0.5, min_intensity=1, norm_order=1)



import warnings

import pytest
from pkg_resources import resource_filename

from .util import SpaceTxValidator
from .validate_sptx import validate

package_name = "sptx_format"
experiment_schema_path = resource_filename(package_name, "schema/experiment.json")
validator = SpaceTxValidator(experiment_schema_path)
example = resource_filename(package_name, "examples/experiment/experiment.json")


def test_fov():
    assert validator.validate_file(example)


def test_nuclei_must_be_present():
    wrong_nuclei = validator.load_json(example)
    wrong_nuclei['auxiliary_images'] = {'not_nuclei': 'nuclei.json'}
    with pytest.warns(UserWarning):
        assert not validator.validate_object(wrong_nuclei)


def test_version_must_be_semantic():
    wrong_version = validator.load_json(example)
    wrong_version['version'] = '10a'
    with pytest.warns(UserWarning):
        assert not validator.validate_object(wrong_version)


def test_dartfish_example_experiment():
    dartfish_example = resource_filename(
        package_name, "examples/experiment/dartfish_experiment.json")
    assert validator.validate_file(dartfish_example)


# see #614
def test_no_manifest_example_experiment():
    no_manifest_example = resource_filename(
        package_name, "examples/experiment/no_fov_manifest.json")

    # SpaceTxValidator doesn't handle multiple files so this passes
    assert validator.validate_file(no_manifest_example)

    # But our tree walker *does* handle multiple files
    with warnings.catch_warnings(record=True) as warnings_:
        assert not validate(no_manifest_example)
        messages = "\n".join([str(x.message) for x in warnings_])
        assert "required property" in messages, f"missing text in '{messages}'"



"""
Tests for codebook._approximate_nearest_code method
"""

import numpy as np

from .test_metric_decode import codebook_factory, intensity_table_factory


def test_simple_intensities_find_correct_nearest_code():
    """
    Test four simple examples for correct decoding. Here the first example should decode to GENE_A,
    the second to GENE_B. The third is closer to GENE_A. The fourth is equidistant to GENE_A and
    GENE_B, but it picks GENE_A because GENE_A comes first in the codebook.
    """
    data = np.array(
        [[[0, 0.5],
          [0.5, 0]],
         [[0, 0.5],
          [0, 0.5]],
         [[0.5, 0],
          [0.5, 0]],
         [[0, 0],
          [0.5, 0.5]]]
    )
    intensities = intensity_table_factory(data=data)
    codebook = codebook_factory()
    distances, gene_ids = codebook._approximate_nearest_code(
        codebook,
        intensities,
        metric='euclidean'
    )

    assert np.array_equal(gene_ids, ['GENE_A', 'GENE_B', 'GENE_A', 'GENE_A'])



import click
import numpy as np

from starfish.intensity_table.intensity_table import IntensityTable
from starfish.types import Features, Indices
from ._base import TargetAssignmentAlgorithm


class Label(TargetAssignmentAlgorithm):

    def __init__(self, **kwargs) -> None:
        """
        Label accepts no parameters, but all pipeline components must accept arbitrary kwargs
        """

    @classmethod
    def _add_arguments(cls, parser) -> None:
        pass

    @staticmethod
    def _assign(
        label_image: np.ndarray,
        intensities: IntensityTable,
        in_place: bool,
    ) -> IntensityTable:

        if len(label_image.shape) == 3:
            cell_ids = label_image[
                intensities[Indices.Z.value].values,
                intensities[Indices.Y.value].values,
                intensities[Indices.X.value].values
            ]
        elif len(label_image.shape) == 2:
            cell_ids = label_image[
                intensities[Indices.Y.value].values,
                intensities[Indices.X.value].values
            ]
        else:
            raise ValueError(
                f"`label_image` must be 2 or 3 dimensional, not {len(label_image.shape)}D."
            )

        if not in_place:
            intensities = intensities.copy()

        intensities[Features.CELL_ID] = (Features.AXIS, cell_ids)

        return intensities

    def run(
            self, label_image: np.ndarray, intensity_table: IntensityTable, verbose: bool=False,
            in_place: bool=False,
    ) -> IntensityTable:
        """Extract cell ids for features in IntensityTable from a segmentation label image

        Parameters
        ----------
        label_image : np.ndarray[np.uint32]
            integer array produced from segmentation where each pixel in a cell is labeled by the
            same integer, and each cell is labeled by a different integer
        intensities : IntensityTable
            spot information

        Returns
        -------
        IntensityTable :
            IntensityTable with added features variable containing cell ids. Points outside of
            cells will be assigned zero.

        """
        return self._assign(label_image, intensity_table, in_place=in_place)

    @staticmethod
    @click.command("Label")
    @click.pass_context
    def _cli(ctx):
        ctx.obj["component"]._cli_run(ctx, Label())



import io

from pkg_resources import resource_filename

from starfish.codebook._format import CURRENT_VERSION, DocumentKeys
from .util import Fuzzer, SpaceTxValidator

package_name = "sptx_format"
codebook_schema_path = resource_filename(package_name, "schema/codebook/codebook.json")
experiment_schema_path = resource_filename(package_name, "schema/experiment.json")
codebook_validator = SpaceTxValidator(codebook_schema_path)
experiment_validator = SpaceTxValidator(experiment_schema_path)


def test_fuzz_mock():
    """
    Starting from the simple obj test class, the Fuzzer will
    propose all the different mutations contained in the the
    values list. The mocked validator will pop them off, in
    order, and compare what it is being given. For testing
    purposes, it will always return true.
    """
    obj = {
        "list": [1, 2, 3]
    }
    values = [
        {"list": [1, 2, 3], "fake": "!"},
        {},
        {"list": 123456789},
        {"list": "fake"},
        {"list": {}},
        {"list": []},
        {"list": [1, 2, 3, '!']},
        {"list": [2, 3]},
        {"list": [123456789, 2, 3]},
        {"list": ["fake", 2, 3]},
        {"list": [{}, 2, 3]},
        {"list": [[], 2, 3]},
        {"list": [1, 2, 3, "!"]},
        {"list": [1, 3]},
        {"list": [1, 123456789, 3]},
        {"list": [1, "fake", 3]},
        {"list": [1, {}, 3]},
        {"list": [1, [], 3]},
        {"list": [1, 2, 3, '!']},
        {"list": [1, 2]},
        {"list": [1, 2, 123456789]},
        {"list": [1, 2, "fake"]},
        {"list": [1, 2, {}]},
        {"list": [1, 2, []]},
    ]

    class Validator(object):
        def __init__(self):
            self.called = 0

        def is_valid(self, obj):
            self.called += 1
            assert obj == values.pop(0), self.called
            return True

    Fuzzer(Validator(), obj).fuzz()


def test_fuzz_codebook():
    """
    simple validation of a hard-coded example codebook.

    The actual values don't matter overly much, but it
    provides a good example of the output.
    """
    codebook = {
        DocumentKeys.VERSION_KEY: str(CURRENT_VERSION),
        DocumentKeys.MAPPINGS_KEY: [
            {
                "codeword": [
                    {"r": 0, "c": 0, "v": 1},
                    {"r": 0, "c": 1, "v": 1},
                    {"r": 1, "c": 1, "v": 1}
                ],
                "target": "SCUBE2"
            }
        ]
    }
    out = io.StringIO()
    codebook_validator.fuzz_object(codebook, out=out)
    expect = """> Fuzzing unknown...
A D I S M L	If the letter is present, mutation is valid!
-----------	--------------------------------------------
. . . . . .	version:
. . . . . .	   0.0.0
. . . . . .	mappings:
A . . . . .	  - codeword:
A D I . . .	    - r:
A D I . . .	         0
A . I . . .	      c:
A . I . . .	         0
A D . . . .	      v:
A D . . . .	         1
A D I . . .	    - r:
A D I . . .	         0
A . I . . .	      c:
A . I . . .	         1
A D . . . .	      v:
A D . . . .	         1
A D I . . .	    - r:
A D I . . .	         1
A . I . . .	      c:
A . I . . .	         1
A D . . . .	      v:
A D . . . .	         1
A . . S . .	    target:
A . . S . .	       SCUBE2
"""
    assert expect == out.getvalue()


def test_fuzz_experiment():
    """
    simple validation of a hard-coded example experiment.

    The actual values don't matter overly much, but it
    provides a good example of the output.
    """
    experiment = {
        "version": "0.0.0",
        "images": {
            "primary": "primary_images.json",
            "nuclei": "nuclei.json",
        },
        "codebook": "codebook.json",
        "extras": {
            "is_space_tx_cool": True
        }
    }
    out = io.StringIO()
    experiment_validator.fuzz_object(experiment, out=out)
    expect = """> Fuzzing unknown...
A D I S M L	If the letter is present, mutation is valid!
-----------	--------------------------------------------
. . . . . .	version:
. . . . . .	   0.0.0
. . . . . .	images:
. . . . . .	   primary:
. . . . . .	      primary_images.json
. D . . . .	   nuclei:
. D . . . .	      nuclei.json
. . . . . .	codebook:
. . . . . .	   codebook.json
. D I S M L	extras:
A D I S M L	   is_space_tx_cool:
A D I S M L	      True
"""
    assert expect == out.getvalue()



import os
import unittest

from starfish.util import exec


class TestWithBuildData(unittest.TestCase):

    STAGES = (
        [
            "starfish", "build",
            "--fov-count=2", '--hybridization-dimensions={"z": 3}',
            lambda tempdir: tempdir
        ],
        [
            "starfish", "validate", "--experiment-json",
            lambda tempdir: os.sep.join([tempdir, "experiment.json"])
        ],
    )

    def test_run_build(self):
        exec.stages(
            TestWithBuildData.STAGES,
            keep_data=("TEST_BUILD_KEEP_DATA" in os.environ))



"""
Tests for IntensityTable.from_spot_data method
"""

import numpy as np
import pandas as pd
import pytest

from starfish import IntensityTable
from starfish.types import Features, Indices, SpotAttributes


def spot_attribute_factory(n: int) -> SpotAttributes:
    """
    Construct SpotAttributes with n synthetic attributes. Each attribute has radius 1 and
    x, y, z coordinates equal to their index i in [0, n)
    """
    return SpotAttributes(
        pd.DataFrame(
            data=np.array([[i, i, i, 1] for i in np.arange(n)]),
            columns=[Indices.Z, Indices.Y, Indices.X, Features.SPOT_RADIUS]
        )
    )


def test_intensity_table_can_be_constructed_from_a_numpy_array_and_spot_attributes():
    """
    Verify that the IntensityTable can be created and that the resulting data matches the array
    it was constructed from.
    """
    spot_attributes = spot_attribute_factory(3)
    data = np.zeros(30).reshape(3, 5, 2)
    intensities = IntensityTable.from_spot_data(data, spot_attributes)

    assert intensities.shape == data.shape
    assert np.array_equal(intensities.values, data)


def test_from_spot_attributes_must_have_aligned_dimensions_spot_attributes_and_data():
    """
    Number of features must match number of SpotAttributes. Pass two attributes and 3 features and
    verify a ValueError is raised.
    """
    spot_attributes = spot_attribute_factory(2)
    data = np.zeros(30).reshape(3, 5, 2)
    with pytest.raises(ValueError):
        IntensityTable.from_spot_data(data, spot_attributes)



from starfish import Experiment


def MERFISH(use_test_data: bool=False):
    if use_test_data:
        return Experiment.from_json(
            'https://d2nhj9g34unfro.cloudfront.net/20181005/MERFISH-TEST/experiment.json')
    return Experiment.from_json(
        'https://d2nhj9g34unfro.cloudfront.net/20181005/MERFISH/experiment.json')


def allen_smFISH(use_test_data: bool=False):
    return Experiment.from_json(
        'https://d2nhj9g34unfro.cloudfront.net/20181005/allen_smFISH/experiment.json')


def DARTFISH(use_test_data: bool=False):
    if use_test_data:
        return Experiment.from_json(
            'https://d2nhj9g34unfro.cloudfront.net/20181005/DARTFISH-TEST/experiment.json')
    return Experiment.from_json(
        'https://d2nhj9g34unfro.cloudfront.net/20181005/DARTFISH/experiment.json')


def ISS(use_test_data: bool=False):
    if use_test_data:
        return Experiment.from_json(
            'https://d2nhj9g34unfro.cloudfront.net/20181005/ISS-TEST/experiment.json')
    return Experiment.from_json(
        'https://d2nhj9g34unfro.cloudfront.net/20181005/ISS/experiment.json')


def osmFISH(use_test_data: bool=False):
    """Return osmFISH data from Codeluppi et al. 2018

    This function returns a single round of a single field of view from the 16 field of view study
    if use_test_data is True, or three fields of view containing all rounds of data if use_test_data
    is False.

    Parameters
    ----------
    use_test_data : bool
        If True, return one round from one field of view, suitable for testing (default False)

    Notes
    -----
    - osmFISH fields of view are quite large (14, 2, 45, 2048, 2048) which takes up approximately
      21 gb in memory. Use the non-test data with care.

    See Also
    --------
    Codeluppi et al. 2018: https://www.nature.com/articles/s41592-018-0175-z
    """
    if use_test_data:
        return Experiment.from_json(
            'https://d2nhj9g34unfro.cloudfront.net/20181005/osmFISH/experiment.json')
    return Experiment.from_json(
        "https://d2nhj9g34unfro.cloudfront.net/20181031/osmFISH/experiment.json")


def BaristaSeq(use_test_data: bool=False) -> Experiment:
    """Loads a BaristaSeq dataset generated from mouse visual cortex. The extracted field of view
    comes from an internal layer of V1 (range: 2-5)

    Parameters
    ----------
    use_test_data : bool
        This parameter is not used for this data type, as there is no data of testing size.

    Returns
    -------
    Experiment
        Experiment containing raw image data
    """
    return Experiment.from_json(
        "https://d2nhj9g34unfro.cloudfront.net/browse/formatted/20181028/"
        "BaristaSeq/cropped_formatted/experiment.json"
    )


def ImagingMassCytometry(use_test_data: bool=False):
    return Experiment.from_json(
        "https://d2nhj9g34unfro.cloudfront.net/browse/formatted/20181023/"
        "imaging_cytof/BodenmillerBreastCancerSamples/experiment.json"
    )





class AlgorithmBase:
    @classmethod
    def _get_algorithm_name(cls):
        """
        Returns the name of the algorithm.  This should be a valid python identifier, i.e.,
        https://docs.python.org/3/reference/lexical_analysis.html#identifiers
        """
        return cls.__name__



from functools import partial
from typing import Optional

import click
import numpy as np

from starfish.imagestack.imagestack import ImageStack
from ._base import FilterAlgorithmBase
from .util import determine_axes_to_group_by


class Clip(FilterAlgorithmBase):

    def __init__(self, p_min: int=0, p_max: int=100, is_volume: bool=False) -> None:
        """Image clipping filter

        Parameters
        ----------
        p_min : int
            values below this percentile are set to p_min (default 0)
        p_max : int
            values above this percentile are set to p_max (default 100)
        is_volume : bool
            If True, 3d (z, y, x) volumes will be filtered. By default, filter 2-d (y, x) tiles
        """
        self.p_min = p_min
        self.p_max = p_max
        self.is_volume = is_volume

    _DEFAULT_TESTING_PARAMETERS = {"p_min": 0, "p_max": 100}

    @staticmethod
    def _clip(image: np.ndarray, p_min: int, p_max: int) -> np.ndarray:
        """Clip values of img below and above percentiles p_min and p_max

        Parameters
        ----------
        image : np.ndarray
            image to be clipped
        p_min : int
          values below this percentile are set to the value of this percentile
        p_max : int
          values above this percentile are set to the value of this percentile

        Notes
        -----
        - Wrapper for np.clip
        - No shifting or transformation to adjust dynamic range is done after clipping

        Returns
        -------
        np.ndarray :
          Numpy array of same shape as img

        """
        v_min, v_max = np.percentile(image, [p_min, p_max])

        return image.clip(min=v_min, max=v_max)

    def run(
            self, stack: ImageStack, in_place: bool=False, verbose: bool=False,
            n_processes: Optional[int]=None
    ) -> ImageStack:
        """Perform filtering of an image stack

        Parameters
        ----------
        stack : ImageStack
            Stack to be filtered.
        in_place : bool
            if True, process ImageStack in-place, otherwise return a new stack
        verbose : bool
            If True, report on the percentage completed (default = False) during processing
        n_processes : Optional[int]
            Number of parallel processes to devote to calculating the filter

        Returns
        -------
        ImageStack :
            If in-place is False, return the results of filter as a new stack.  Otherwise return the
            original stack.

        """
        group_by = determine_axes_to_group_by(self.is_volume)
        clip = partial(self._clip, p_min=self.p_min, p_max=self.p_max)
        result = stack.apply(
            clip,
            group_by=group_by, verbose=verbose, in_place=in_place, n_processes=n_processes
        )
        return result

    @staticmethod
    @click.command("Clip")
    @click.option(
        "--p-min", default=0, type=int, help="clip intensities below this percentile")
    @click.option(
        "--p-max", default=100, type=int, help="clip intensities above this percentile")
    @click.pass_context
    def _cli(ctx, p_min, p_max):
        ctx.obj["component"]._cli_run(ctx, Clip(p_min, p_max))



import collections
import multiprocessing
import os
import warnings
from copy import deepcopy
from functools import partial
from itertools import product
from typing import (
    Any,
    Callable,
    Iterator,
    List,
    Mapping,
    MutableSequence,
    Optional,
    Sequence,
    Set,
    Tuple,
    Union,
)

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import xarray as xr
from matplotlib import get_backend as get_matplotlib_backend
from scipy.ndimage.filters import gaussian_filter
from scipy.stats import scoreatpercentile
from skimage import exposure
from skimage import img_as_float32, img_as_uint
from slicedimage import Reader, TileSet, Writer
from slicedimage.io import resolve_path_or_url
from tqdm import tqdm

from starfish.errors import DataFormatWarning
from starfish.experiment.builder import build_image, TileFetcher
from starfish.experiment.builder.defaultproviders import OnesTile, tile_fetcher_factory
from starfish.imagestack import indexing_utils
from starfish.imagestack import physical_coordinate_calculator
from starfish.intensity_table.intensity_table import IntensityTable
from starfish.multiprocessing.shmem import SharedMemory
from starfish.types import (
    Coordinates,
    Indices,
    PHYSICAL_COORDINATE_DIMENSION,
    PhysicalCoordinateTypes,
)
from starfish.util import StripArguments
from ._mp_dataarray import MPDataArray
from .dataorder import AXES_DATA, N_AXES


class ImageStack:
    """
    Container for a TileSet (field of view)

    Attributes
    ----------
    num_chs : int
        the number of channels stored in the image tensor
    num_rounds : int
        the number of imaging rounds stored in the image tensor
    num_zlayers : int
        the number of z-layers stored in the image tensor
    numpy_array : np.ndarray
        the 5-d image tensor is stored in this array
    raw_shape : Tuple[int]
        the shape of the image tensor (in integers)
    shape : Dict[str, int]
           the shape of the image tensor by categorical index (channels, imaging rounds, z-layers)

    Methods
    -------
    get_slice(indices)
        retrieve a slice of the image tensor
    set_slice(indices, data, axes=None)
        set a slice of the image tensor
    apply(func, group_by={Indices.ROUND, Indices.CH, Indices.Z}, in_place=False, verbose=False,
            n_processes=None)
        split the image tensor along one or more axes and apply a function across each of the
        components to yield an image tensor
    transform(func, group_by={Indices.ROUND, Indices.CH, Indices.Z}, verbose=False,
            n_processes=None)
        split the image tensor along one or more axes and apply a function across each of the
        components. Results are returned as a List with length equal to the number of times
        the image tensor is split.
    max_proj(*dims)
        return a max projection over one or more axis of the image tensor
    show_stack(indices, color_map='gray', figure_size=(10, 10), rescale=False, p_min=None,
            p_max=None)
        show an interactive, pageable view of the image tensor, or a slice of the image tensor
    write(filepath, tile_opener=None)
        save the (potentially modified) image tensor to disk
    """

    def __init__(self, image_partition: TileSet) -> None:
        self._image_partition = image_partition
        self._tile_shape = image_partition.default_tile_shape

        # Examine the tiles to figure out the right kind (int, float, etc.) and size.  We require
        # that all the tiles have the same kind of data type, but we do not require that they all
        # have the same size of data type. The # allocated array is the highest size we encounter.
        kind = None
        max_size = 0
        for tile in tqdm(self._image_partition.tiles()):
            dtype = tile.numpy_array.dtype
            if kind is None:
                kind = dtype.kind
            else:
                if kind != dtype.kind:
                    raise TypeError("All tiles should have the same kind of dtype")
            if dtype.itemsize > max_size:
                max_size = dtype.itemsize
            if self._tile_shape is None:
                self._tile_shape = tile.tile_shape
            elif tile.tile_shape is not None and self._tile_shape != tile.tile_shape:
                raise ValueError("Starfish does not support tiles that are not identical in shape")

        shape: MutableSequence[int] = []
        dims: MutableSequence[str] = []
        coordinates_shape: MutableSequence[int] = []
        coordinates_dimensions: MutableSequence[str] = []
        for ix in range(N_AXES):
            size_for_axis: Optional[int] = None
            dim_for_axis: Optional[Indices] = None

            for axis_name, axis_data in AXES_DATA.items():
                if ix == axis_data.order:
                    size_for_axis = self._get_dimension_size(axis_name)
                    dim_for_axis = axis_name
                    break

            if size_for_axis is None or dim_for_axis is None:
                raise ValueError(
                    f"Could not find entry for the {ix}th axis in AXES_DATA")

            shape.append(size_for_axis)
            dims.append(dim_for_axis.value)
            coordinates_shape.append(size_for_axis)
            coordinates_dimensions.append(dim_for_axis.value)

        shape.extend(self._tile_shape)
        dims.extend([Indices.Y.value, Indices.X.value])
        coordinates_dimensions.append(PHYSICAL_COORDINATE_DIMENSION)
        coordinates_shape.append(6)
        # now that we know the tile data type (kind and size), we can allocate the data array.
        self._data = MPDataArray.from_shape_and_dtype(
            shape=shape,
            dtype=np.float32,
            initial_value=0,
            dims=dims,
        )
        self._coordinates = xr.DataArray(
            np.empty(
                shape=coordinates_shape,
                dtype=np.float32,
            ),
            dims=coordinates_dimensions,
            coords={
                PHYSICAL_COORDINATE_DIMENSION: [
                    PhysicalCoordinateTypes.X_MIN.value,
                    PhysicalCoordinateTypes.X_MAX.value,
                    PhysicalCoordinateTypes.Y_MIN.value,
                    PhysicalCoordinateTypes.Y_MAX.value,
                    PhysicalCoordinateTypes.Z_MIN.value,
                    PhysicalCoordinateTypes.Z_MAX.value,
                ],
            },
        )

        # iterate through the tiles and set the data.
        for tile in self._image_partition.tiles():
            h = tile.indices[Indices.ROUND]
            c = tile.indices[Indices.CH]
            zlayer = tile.indices.get(Indices.Z, 0)
            data = tile.numpy_array

            if max_size != data.dtype.itemsize:
                warnings.warn(
                    f"Tile "
                    f"(R: {tile.indices[Indices.ROUND]} C: {tile.indices[Indices.CH]} "
                    f"Z: {tile.indices[Indices.Z]}) has "
                    f"dtype {data.dtype}.  One or more tiles is of a larger dtype "
                    f"{self._data.dtype}.",
                    DataFormatWarning)

            data = img_as_float32(data)
            self.set_slice(indices={Indices.ROUND: h, Indices.CH: c, Indices.Z: zlayer}, data=data)
            coordinate_selector = {
                Indices.ROUND.value: h,
                Indices.CH.value: c,
                Indices.Z.value: zlayer,
            }
            coordinates_values = [
                tile.coordinates[Coordinates.X][0], tile.coordinates[Coordinates.X][1],
                tile.coordinates[Coordinates.Y][0], tile.coordinates[Coordinates.Y][1],
            ]
            if Coordinates.Z in tile.coordinates:
                coordinates_values.extend([
                    tile.coordinates[Coordinates.Z][0], tile.coordinates[Coordinates.Z][1],
                ])
            else:
                coordinates_values.extend([np.nan, np.nan])

            self._coordinates.loc[coordinate_selector] = np.array(coordinates_values)

    @staticmethod
    def _validate_data_dtype_and_range(data: Union[np.ndarray, xr.DataArray]) -> None:
        """verify that data is of dtype float32 and in range [0, 1]"""
        if data.dtype != np.float32:
            raise TypeError(
                f"ImageStack data must be of type float32, not {data.dtype}. Please convert data "
                f"using skimage.img_as_float32 prior to calling set_slice."
            )
        if np.min(data) < 0 or np.max(data) > 1:
            raise ValueError(
                f"ImageStack data must be of type float32 and in the range [0, 1]. Please convert "
                f"data using skimage.img_as_float32 prior to calling set_slice."
            )

    def __repr__(self):
        shape = ', '.join(f'{k}: {v}' for k, v in self._data.sizes.items())
        return f"<starfish.ImageStack ({shape})>"

    @classmethod
    def from_url(cls, url: str, baseurl: Optional[str]):
        """
        Constructs an ImageStack object from a URL and a base URL.

        The following examples will all load from the same location:
          1. url: https://www.example.com/images/hybridization.json  baseurl: None
          2. url: https://www.example.com/images/hybridization.json  baseurl: I_am_ignored
          3. url: hybridization.json  baseurl: https://www.example.com/images
          4. url: images/hybridization.json  baseurl: https://www.example.com

        Parameters
        ----------
        url : str
            Either an absolute URL or a relative URL referring to the image to be read.
        baseurl : Optional[str]
            If url is a relative URL, then this must be provided.  If url is an absolute URL, then
            this parameter is ignored.
        """
        image_partition = Reader.parse_doc(url, baseurl)

        return cls(image_partition)

    @classmethod
    def from_path_or_url(cls, url_or_path: str) -> "ImageStack":
        """
        Constructs an ImageStack object from an absolute URL or a filesystem path.

        The following examples will all load from the same location:
          1. url_or_path: file:///Users/starfish-user/images/hybridization.json
          2. url_or_path: /Users/starfish-user/images/hybridization.json

        Parameters
        ----------
        url_or_path : str
            Either an absolute URL or a filesystem path to an imagestack.
        """
        _, relativeurl, baseurl = resolve_path_or_url(url_or_path)
        return cls.from_url(relativeurl, baseurl)

    @classmethod
    def from_numpy_array(cls, array: np.ndarray) -> "ImageStack":
        """Create an ImageStack from a 5d numpy array with shape (n_round, n_ch, n_z, y, x)

        Parameters
        ----------
        array : np.ndarray
            5-d tensor of shape (n_round, n_ch, n_z, y, x)

        Returns
        -------
        ImageStack :
            array data stored as an ImageStack

        """
        if len(array.shape) != 5:
            raise ValueError('a 5-d tensor with shape (n_round, n_ch, n_z, y, x) must be provided.')
        try:
            cls._validate_data_dtype_and_range(array)
        except TypeError:
            warnings.warn(f"ImageStack detected as {array.dtype}. Converting to float32...")
            array = img_as_float32(array)

        n_round, n_ch, n_z, height, width = array.shape
        empty = cls.synthetic_stack(
            num_round=n_round, num_ch=n_ch, num_z=n_z, tile_height=height, tile_width=width)

        for h in np.arange(n_round):
            for c in np.arange(n_ch):
                for z in np.arange(n_z):
                    view = array[h, c, z]
                    empty.set_slice({Indices.ROUND: h, Indices.CH: c, Indices.Z: z}, view)

        return empty

    @property
    def xarray(self) -> xr.DataArray:
        """Retrieves the image data as an xarray.DataArray"""
        return self._data.data

    def sel(self, indexers: Mapping[Indices, Union[int]]):
        """Given a dictionary mapping the index name to either a value or a slice range, return an
        Imagestack with each dimension indexed accordingly

        Parameters
        ----------
        indexers : Dict[Indices, (int/tuple)]
            A dictionary of dim:index where index is the value or range to index the dimension

        Examples
        --------

        Create an Imagestack using the ``synthetic_stack`` method::
        >>> from starfish import ImageStack
        >>> from starfish.types import Indices
        >>> stack = ImageStack.synthetic_stack(5, 5, 15, 200, 200)
        >>> stack
        <starfish.ImageStack (r: 5, c: 5, z: 15, y: 200, x: 200)>
        >>> stack.sel({Indices.ROUND: (1, None), Indices.CH: 0, Indices.Z: 0})
        <starfish.ImageStack (r: 4, c: 1, z: 1, y: 200, x: 200)>
        >>> stack.sel({Indices.ROUND: 0, Indices.CH: 0, Indices.Z: 1,
        ...Indices.Y: 100, Indices.X: (None, 100)})
        <starfish.ImageStack (r: 1, c: 1, z: 1, y: 1, x: 100)>
        and the imagestack's physical coordinates
        xarray also indexed and recalculated according to the x,y slicing.

        Returns
        -------
        ImageStack :
            a new image stack indexed by given value or range.
        """

        # convert indexers to Dict[str, (int/slice)] format
        formatted_indexers = indexing_utils.convert_to_indexers_dict(indexers)
        indexed_data = indexing_utils.index_keep_dimensions(self.xarray, formatted_indexers)
        stack = self.from_numpy_array(indexed_data.data)
        # set coords on new stack
        stack._coordinates = physical_coordinate_calculator.calc_new_physical_coords_array(
            self._coordinates, self.shape, formatted_indexers)
        return stack

    def get_slice(
            self,
            indices: Mapping[Indices, Union[int, slice]]
    ) -> Tuple[np.ndarray, Sequence[Indices]]:
        """
        Given a dictionary mapping the index name to either a value or a slice range, return a
        numpy array representing the slice, and a list of the remaining axes beyond the normal x-y
        tile.

        Examples
        --------
        ImageStack axes: H, C, and Z with shape 3, 4, 5, respectively.
        ImageStack Implicit axes: X, Y with shape 10, 20, respectively.
        Called to slice with indices {Z: 5}.
        Result: a 4-dimensional numpy array with shape (3, 4, 20, 10) and the remaining axes [H, C].

        Original axes: H, C, and Z.
        Implicit axes: X, Y.
        Called to slice with indices {Z: 5, C: slice(2, 4)}.
        Result: a 4-dimensional numpy array with shape (3, 2, 20, 10) and the remaining axes [H, C].

        """
        slice_list, axes = self._build_slice_list(indices)
        result = self._data.values[slice_list]

        if result.dtype != np.float32:
            warnings.warn(
                f"Non-float32 dtype: {result.dtype} detected. Data has likely been set using "
                f"private attributes of ImageStack. ImageStack only supports float data in the "
                f"range [0, 1]. Many algorithms will not function properly if provided other "
                f"DataTypes. See: http://scikit-image.org/docs/dev/user_guide/data_types.html")

        return result, axes

    def set_slice(
            self,
            indices: Mapping[Indices, Union[int, slice]],
            data: np.ndarray,
            axes: Sequence[Indices]=None):
        """
        Given a dictionary mapping the index name to either a value or a slice range and a source
        numpy array, set the slice of the array of this ImageStack to the values in the source
        numpy array. If the optional parameter axes is provided, that represents the axes of the
        numpy array beyond the x-y tile.

        Examples
        --------
        ImageStack axes: H, C, and Z with shape 3, 4, 5, respectively.
        ImageStack Implicit axes: X, Y with shape 10, 20, respectively.
        Called to set a slice with indices {Z: 5}.
        Data: a 4-dimensional numpy array with shape (3, 4, 10, 20)
        Result: Replace the data for Z=5.

        ImageStack axes: H, C, and Z. (shape 3, 4, 5)
        ImageStack Implicit axes: X, Y. (shape 10, 20)
        Called to set a slice with indices {Z: 5, C: slice(2, 4)}.
        Data: a 4-dimensional numpy array with shape (3, 2, 10, 20)
        Result: Replace the data for Z=5, C=2-3.
        """

        self._validate_data_dtype_and_range(data)

        slice_list, expected_axes = self._build_slice_list(indices)

        if axes is not None:
            if len(axes) != len(data.shape) - 2:
                raise ValueError(
                    "data shape ({}) should be the axes ({}) and (x,y).".format(data.shape, axes))
            move_src = list()
            move_dst = list()
            for src_idx, axis in enumerate(axes):
                try:
                    dst_idx = expected_axes.index(axis)
                except ValueError:
                    raise ValueError(
                        "Unexpected axis {}.  Expecting only {}.".format(axis, expected_axes))
                if src_idx != dst_idx:
                    move_src.append(src_idx)
                    move_dst.append(dst_idx)

            if len(move_src) != 0:
                data = data.view()
                np.moveaxis(data, move_src, move_dst)

        if self._data[slice_list].shape != data.shape:
            raise ValueError("source shape {} mismatches destination shape {}".format(
                data.shape, self._data[slice_list].shape))

        self._data.values[slice_list] = data

    def show_stack_napari(self, indices: Mapping[Indices, Union[int, slice]]):
        """Displays the image stack using Napari (https://github.com/Napari)

        Parameters
        ----------
        indices : Mapping[Indices, Union[int, slice]],
            Indices to select a volume to visualize. Passed to `Image.get_slice()`.
            See `Image.get_slice()` for examples.

        Notes
        -----
        To use in a Jupyter notebook, use the %gui qt5 magic.
        Axes currently cannot be labeled. Until such a time that they can, this function will
            order them by Round, Channel, and Z.


        """
        try:
            import napari_gui
        except ImportError:
            warnings.warn("Cannot find the napari library. "
                          "Install it by running \"pip install napari\"")
            return
        # TODO ambrosejcarr: this should use updated imagestack slicing routines when they are added
        # and indices should be optional to enable full stack viewing.
        # Switch axes such that it is indexed [x, y, round, channel, z]
        slices, axes = self.get_slice(indices)
        reordered_array = np.moveaxis(slices, [-2, -1], [0, 1])

        napari_gui.imshow(reordered_array, multichannel=False)

    def show_stack(
            self, indices: Mapping[Indices, Union[int, slice]],
            color_map: str= 'gray', figure_size: Tuple[int, int]=(10, 10),
            rescale: bool=False, p_min: Optional[float]=None, p_max: Optional[float]=None, **kwargs
    ):
        """Create an interactive visualization of an image stack

        Produces a slider that flips through the selected volume tile-by-tile. Supports manual
        adjustment of dynamic range.

        Parameters
        ----------
        indices : Mapping[Indices, Union[int, slice]],
            Indices to select a volume to visualize. Passed to `Image.get_slice()`.
            See `Image.get_slice()` for examples.
        color_map : str (default = 'gray')
            string id of a matplotlib colormap
        figure_size : Tuple[int, int] (default = (10, 10))
            size of the figure in inches
        rescale : bool (default = False)
            if True, rescale the data to exclude high and low-value outliers
            (see skimage.exposure.rescale_intensity).
        p_min: float
            clip values below this intensity percentile. If provided, overrides rescale, above.
            (default = None)
        p_max: float
            clip values above this intensity percentile. If provided, overrides rescale, above.
            (default = None)

        Raises
        ------
        ValueError :
            User must select one of rescale or p_min/p_max to adjust the image dynamic range.
            If both are selected, a ValueError is raised.

        Notes
        -----
        For this widget to function interactively in the notebook, after ipywidgets has been
        installed, the user must register the widget with jupyter by typing the following command
        into the terminal: jupyter nbextension enable --py widgetsnbextension

        """

        # infer if %matplotlib inline or notebook
        mpl_is_notebook = 'nbAgg' in get_matplotlib_backend()

        if not indices:
            raise ValueError('indices may not be an empty dict or None')

        # get linearized scaled and clipped tiles, along with title names, for plotting
        linear_view, labels, n_tiles = self._get_scaled_clipped_linear_view(indices,
                                                                            rescale,
                                                                            p_min,
                                                                            p_max
                                                                            )

        if mpl_is_notebook:
            self._show_matplotlib_notebook(
                linear_view,
                labels,
                n_tiles,
                figure_size,
                color_map
            )
        else:
            return self._show_matplotlib_inline(
                linear_view,
                labels,
                n_tiles,
                figure_size,
                color_map
            )

    def _get_scaled_clipped_linear_view(self, indices, rescale, p_min, p_max):

        # get the requested chunk, linearize the remaining data into a sequence of tiles
        data, remaining_inds = self.get_slice(indices)

        # identify the dimensionality of data with all dimensions other than x, y linearized
        if len(data.shape) >= 3:
            n_tiles = np.product(data.shape[:-2])
        else:
            raise ValueError(
                f'a stack with dimensionality >= 3 is required, the provided indexer produced a '
                f'stack with shape {data.shape}')

        # linearize the array
        linear_view: np.ndarray = data.reshape((n_tiles,) + data.shape[-2:])

        # set the labels for the linearized tiles
        labels: List[List[str]] = []
        for index, size in zip(remaining_inds, data.shape[:-2]):
            labels.append([f'{index}{n}' for n in range(size)])

        # mypy thinks this has an incompatible type "Iterator[Tuple[Any, ...]]";
        # it expects "Iterable[List[str]]"
        labels = list(product(*labels))  # type: ignore

        n_tiles = linear_view.shape[0]

        if rescale and any((p_min, p_max)):
            raise ValueError('select one of rescale and p_min/p_max to rescale image, not both.')

        elif rescale:
            print("Rescaling ...")
            vmin, vmax = scoreatpercentile(data, (0.5, 99.5))
            linear_view = exposure.rescale_intensity(
                linear_view,
                in_range=(vmin, vmax),
                out_range=np.float32
            ).astype(np.float32)

        elif p_min or p_max:
            print("Clipping ...")
            a_min, a_max = scoreatpercentile(
                linear_view,
                (p_min if p_min else 0, p_max if p_max else 100)
            )
            linear_view = np.clip(linear_view, a_min=a_min, a_max=a_max)

        return linear_view, labels, n_tiles

    @staticmethod
    def _show_matplotlib_notebook(
            linear_view, labels, n_tiles, figure_size, color_map
    ):
        from ipywidgets import interact, fixed

        fig, ax = plt.subplots(figsize=figure_size)
        im = ax.imshow(linear_view[0], cmap=color_map)
        ax.set_xticks([])
        ax.set_yticks([])

        def show_plane(ax, plane, plane_index, cmap="gray", title=None):
            # Update the image in the current plane
            im.set_data(plane)
            if title:
                ax.set_title(title)

        def display_slice(plane_index, ax):
            title_str = " ".join(str(lab).upper() for lab in labels[plane_index])
            show_plane(ax, linear_view[plane_index], plane_index, title=title_str, cmap=color_map)

        interact(display_slice, ax=fixed(ax), plane_index=(0, n_tiles - 1))

    @staticmethod
    def _show_matplotlib_inline(
            linear_view, labels, n_tiles, figure_size, color_map
    ):
        from ipywidgets import interact

        def show_plane(ax, plane, plane_index, cmap="gray", title=None):
            ax.imshow(plane, cmap=cmap)

            if title:
                ax.set_title(title, fontsize=16)
            ax.set_xticks([])
            ax.set_yticks([])

        @interact(plane_index=(0, n_tiles - 1))
        def display_slice(plane_index=0):
            fig, ax = plt.subplots(figsize=figure_size)
            title_str = " ".join(str(lab).upper() for lab in labels[plane_index])
            show_plane(ax, linear_view[plane_index], plane_index, title=title_str, cmap=color_map)
            plt.show()

        return display_slice

    @staticmethod
    def _build_slice_list(
            indices: Mapping[Indices, Union[int, slice]]
    ) -> Tuple[Tuple[Union[int, slice], ...], Sequence[Indices]]:
        slice_list: MutableSequence[Union[int, slice]] = [
            slice(None, None, None)
            for _ in range(N_AXES)
        ]
        axes = []
        removed_axes = set()
        for name, value in indices.items():
            idx = AXES_DATA[name].order
            if not isinstance(value, slice):
                removed_axes.add(name)
            slice_list[idx] = value

        for dimension_value, dimension_name in sorted([
            (dimension_value.order, dimension_name)
            for dimension_name, dimension_value in AXES_DATA.items()
        ]):
            if dimension_name not in removed_axes:
                axes.append(dimension_name)

        return tuple(slice_list), axes

    def _iter_indices(self, indices: Set[Indices]={Indices.ROUND, Indices.CH}
                      ) -> Iterator[Mapping[Indices, int]]:
        """Iterate over provided indices

        Parameters
        ----------
        indices : Set[Indices]
            The set of Indices to be iterated over

        Yields
        ------
        Dict[str, int]
            Mapping of dimension name to index

        """
        ranges = [np.arange(self.shape[ind]) for ind in indices]
        for items in product(*ranges):
            a = zip(indices, items)
            yield {ind: val for (ind, val) in a}

    def apply(
            self,
            func: Callable,
            group_by: Set[Indices]=None,
            in_place=False,
            verbose: bool=False,
            n_processes: Optional[int]=None,
            **kwargs
    ) -> "ImageStack":
        """Split the image along a set of axes and apply a function across all the components.  This
        function should yield data of the same dimensionality as the input components.  These
        resulting components are then constituted into an ImageStack and returned.

        Parameters
        ----------
        func : Callable
            Function to apply. must expect a first argument which is a numpy array (see group_by)
            but may return any object type.
        group_by : Set[Indices]
            Axes to split the data along.  For instance, splitting a 2D array (axes: X, Y; size:
            3, 4) by X results in 3 arrays of size 4.  (default {Indices.ROUND, Indices.CH,
            Indices.Z})
        in_place : bool
            (default True) If True, function is executed in place. If n_proc is not 1, the tile or
            volume will be copied once during execution. If false, a new ImageStack object will be
            produced.
        verbose : bool
            If True, report on the percentage completed (default = False) during processing
        n_processes : Optional[int]
            The number of processes to use for apply. If None, uses the output of os.cpu_count()
            (default = None).
        kwargs : dict
            Additional arguments to pass to func

        Returns
        -------
        ImageStack :
            If inplace is False, return a new ImageStack, otherwise return a reference to the
            original stack with data modified by application of func
        """
        if group_by is None:
            group_by = {Indices.ROUND, Indices.CH, Indices.Z}

        if not in_place:
            image_stack = deepcopy(self)
            return image_stack.apply(
                func,
                group_by=group_by, in_place=True, verbose=verbose, n_processes=n_processes, **kwargs
            )
        bound_func = partial(ImageStack._in_place_apply, func)

        self.transform(
            bound_func,
            group_by=group_by,
            verbose=verbose,
            n_processes=n_processes,
            **kwargs)

        return self

    @staticmethod
    def _in_place_apply(apply_func: Callable[..., np.ndarray], data: np.ndarray, **kwargs) -> None:
        result = apply_func(data, **kwargs)
        data[:] = result

    def transform(
            self,
            func: Callable,
            group_by: Set[Indices]=None,
            verbose=False,
            n_processes: Optional[int]=None,
            **kwargs
    ) -> List[Any]:
        """Split the image along a set of axes, and apply a function across all the components.

        Parameters
        ----------
        func : Callable
            Function to apply. must expect a first argument which is a numpy array (see group_by)
            but may return any object type.
        group_by : Set[Indices]
            Axes to split the data along.  For instance, splitting a 2D array (axes: X, Y; size:
            3, 4) by X results in 3 arrays of size 4.  (default {Indices.ROUND, Indices.CH,
            Indices.Z})
        verbose : bool
            If True, report on the percentage completed (default = False) during processing
        n_processes : Optional[int]
            The number of processes to use for apply. If None, uses the output of os.cpu_count()
            (default = None).
        kwargs : dict
            Additional arguments to pass to func being applied

        Returns
        -------
        List[Any] :
            The results of applying func to stored image data
        """
        if group_by is None:
            group_by = {Indices.X, Indices.Y}

        indices = list(self._iter_indices(group_by))
        slice_list = [self._build_slice_list(index)[0]
                      for index in indices]

        indices_and_slice_list = zip(indices, slice_list)
        if verbose:
            indices_and_slice_list = tqdm(indices_and_slice_list)

        applyfunc: Callable = partial(
            self._multiprocessing_workflow,
            StripArguments(partial(func, **kwargs), positional_arguments_removed=(1, 2)),
        )

        with multiprocessing.Pool(
                n_processes,
                initializer=SharedMemory.initializer,
                initargs=((self._data._backing_mp_array,
                           self._data._data.shape,
                           self._data._data.dtype),)) as pool:
            results = pool.imap(applyfunc, indices_and_slice_list)
            return list(zip(results, indices))

    @staticmethod
    def _multiprocessing_workflow(
            worker_callable: Callable[[np.ndarray,
                                       Mapping[Indices, int],
                                       Tuple[Union[int, slice], ...]], Any],
            indices_and_slice_list: Tuple[Mapping[Indices, int],
                                          Tuple[Union[int, slice], ...]],
    ):
        backing_mp_array, shape, dtype = SharedMemory.get_payload()
        unshaped_numpy_array = np.frombuffer(backing_mp_array.get_obj(), dtype=dtype)
        numpy_array = unshaped_numpy_array.reshape(shape)

        sliced = numpy_array[indices_and_slice_list[1]]

        return worker_callable(sliced, *indices_and_slice_list)

    @property
    def tile_metadata(self) -> pd.DataFrame:
        """return a table containing Tile metadata

        Returns
        -------
        pd.DataFrame :
            dataframe containing per-tile metadata information for each image. Guaranteed to
            include information on channel, imaging round, z_layer, and barcode index. Also
            contains any information stored in the extras field for each tile in hybridization.json

        """

        data: collections.defaultdict = collections.defaultdict(list)
        index_keys = set(
            key
            for tile in self._image_partition.tiles()
            for key in tile.indices.keys())
        extras_keys = set(
            key
            for tile in self._image_partition.tiles()
            for key in tile.extras.keys())
        duplicate_keys = index_keys.intersection(extras_keys)
        if len(duplicate_keys) > 0:
            duplicate_keys_str = ", ".join([str(key) for key in duplicate_keys])
            raise ValueError(
                f"keys ({duplicate_keys_str}) was found in both the Tile specification and extras "
                f"field. Tile specification keys may not be duplicated in the extras field.")

        for tile in self._image_partition.tiles():
            for k in index_keys:
                data[k].append(tile.indices.get(k, None))
            for k in extras_keys:
                data[k].append(tile.extras.get(k, None))

            if 'barcode_index' not in tile.extras:
                round_ = tile.indices[Indices.ROUND]
                ch = tile.indices[Indices.CH]
                z = tile.indices.get(Indices.Z, 0)
                barcode_index = (((z * self.num_rounds) + round_) * self.num_chs) + ch

                data['barcode_index'].append(barcode_index)

        return pd.DataFrame(data)

    @property
    def raw_shape(self) -> Tuple[int, int, int, int, int]:
        """
        Returns the shape of the 5-d image tensor stored as self.image

        Returns
        -------
        Tuple[int, int, int, int, int] :
            The size of the image tensor
        """
        return self._data.shape

    @property
    def shape(self) -> collections.OrderedDict:
        """
        Returns the shape of the space that this image inhabits.  It does not include the
        dimensions of the image itself.  For instance, if this is an X-Y image in a C-H-Y-X space,
        then the shape would include the dimensions C and H.

        Returns
        -------
        An ordered mapping between index names to the size of the index.
        """
        # TODO: (ttung) Note that the return type should be ..OrderedDict[Any, str], but python3.6
        # has a bug where this # breaks horribly.  Can't find a bug id to link to, but see
        # https://stackoverflow.com/questions/41207128/how-do-i-specify-ordereddict-k-v-types-for-\
        # mypy-type-annotation
        result: collections.OrderedDict[Any, str] = collections.OrderedDict()
        for name, data in AXES_DATA.items():
            result[name] = self._data.shape[data.order]
        result['y'] = self._data.shape[-2]
        result['x'] = self._data.shape[-1]

        return result

    def coordinates(
            self,
            indices: Mapping[Indices, int],
            physical_axis: Coordinates) -> Tuple[float, float]:
        """Given a set of indices that uniquely identify a tile and a physical axis, return the min
        and the max coordinates for that tile along that axis."""

        return physical_coordinate_calculator.get_coordinates(coords_array=self._coordinates,
                                                              indices=indices,
                                                              physical_axis=physical_axis)

    def _get_dimension_size(self, dimension: Indices):
        axis_data = AXES_DATA[dimension]
        if dimension in self._image_partition.dimensions or axis_data.required:
            return self._image_partition.get_dimension_shape(dimension)
        return 1

    @property
    def num_rounds(self):
        return self._get_dimension_size(Indices.ROUND)

    @property
    def num_chs(self):
        return self._get_dimension_size(Indices.CH)

    @property
    def num_zlayers(self):
        return self._get_dimension_size(Indices.Z)

    @property
    def tile_shape(self):
        return self._tile_shape

    def write(self, filepath: str, tile_opener=None) -> None:
        """write the image tensor to disk in spaceTx format

        Parameters
        ----------
        filepath : str
            Path + prefix for the images and primary_images.json written by this function
        tile_opener : TODO ttung: doc me.

        """
        for tile in self._image_partition.tiles():
            h = tile.indices[Indices.ROUND]
            c = tile.indices[Indices.CH]
            zlayer = tile.indices.get(Indices.Z, 0)
            tile.numpy_array, axes = self.get_slice(
                indices={Indices.ROUND: h, Indices.CH: c, Indices.Z: zlayer}
            )
            assert len(axes) == 0

        seen_x_coords, seen_y_coords, seen_z_coords = set(), set(), set()
        for tile in self._image_partition.tiles():
            seen_x_coords.add(tile.coordinates[Coordinates.X])
            seen_y_coords.add(tile.coordinates[Coordinates.Y])
            z_coords = tile.coordinates.get(Coordinates.Z, None)
            if z_coords is not None:
                seen_z_coords.add(z_coords)

        sorted_x_coords = sorted(seen_x_coords)
        sorted_y_coords = sorted(seen_y_coords)
        sorted_z_coords = sorted(seen_z_coords)
        x_coords_to_idx = {coords: idx for idx, coords in enumerate(sorted_x_coords)}
        y_coords_to_idx = {coords: idx for idx, coords in enumerate(sorted_y_coords)}
        z_coords_to_idx = {coords: idx for idx, coords in enumerate(sorted_z_coords)}

        if tile_opener is None:
            def tile_opener(tileset_path, tile, ext):
                tile_basename = os.path.splitext(tileset_path)[0]
                xcoord = tile.coordinates[Coordinates.X]
                ycoord = tile.coordinates[Coordinates.Y]
                zcoord = tile.coordinates.get(Coordinates.Z, None)
                xcoord = tuple(xcoord) if isinstance(xcoord, list) else xcoord
                ycoord = tuple(ycoord) if isinstance(ycoord, list) else ycoord
                xval = x_coords_to_idx[xcoord]
                yval = y_coords_to_idx[ycoord]
                if zcoord is not None:
                    zval = z_coords_to_idx[zcoord]
                    zstr = "-Z{}".format(zval)
                else:
                    zstr = ""
                return open(
                    "{}-X{}-Y{}{}-H{}-C{}.{}".format(
                        tile_basename,
                        xval,
                        yval,
                        zstr,
                        tile.indices[Indices.ROUND],
                        tile.indices[Indices.CH],
                        ext,
                    ),
                    "wb")

        if not filepath.endswith('.json'):
            filepath += '.json'
        Writer.write_to_path(
            self._image_partition,
            filepath,
            pretty=True,
            tile_opener=tile_opener)

    def max_proj(self, *dims: Indices) -> "ImageStack":
        """return a max projection over one or more axis of the image tensor

        Parameters
        ----------
        dims : Indices
            one or more axes to project over

        Returns
        -------
        np.ndarray :
            max projection

        """
        max_projection = self._data.max([dim.value for dim in dims])
        max_projection = max_projection.expand_dims(tuple(dim.value for dim in dims))
        max_projection = max_projection.transpose(*self.xarray.dims)
        max_proj_stack = self.from_numpy_array(max_projection.values)
        return max_proj_stack

    def _squeezed_numpy(self, *dims: Indices):
        """return this ImageStack's data as a squeezed numpy array"""
        return self.xarray.squeeze(tuple(dim.value for dim in dims)).values

    @classmethod
    def synthetic_stack(
            cls,
            num_round: int=4,
            num_ch: int=4,
            num_z: int=12,
            tile_height: int=50,
            tile_width: int=40,
            tile_fetcher: TileFetcher=None,
    ) -> "ImageStack":
        """generate a synthetic ImageStack

        Returns
        -------
        ImageStack :
            imagestack containing a tensor whose default shape is (2, 3, 4, 30, 20)
            and whose default values are all 1.

        """
        if tile_fetcher is None:
            tile_fetcher = tile_fetcher_factory(
                OnesTile,
                False,
                (tile_height, tile_width),
            )

        collection = build_image(
            1,
            num_round,
            num_ch,
            num_z,
            tile_fetcher,
        )
        tileset = list(collection.all_tilesets())[0][1]

        return ImageStack(tileset)

    @classmethod
    def synthetic_spots(
            cls, intensities: IntensityTable, num_z: int, height: int, width: int,
            n_photons_background=1000, point_spread_function=(4, 2, 2),
            camera_detection_efficiency=0.25, background_electrons=1,
            graylevel: float=37000.0 / 2 ** 16, ad_conversion_bits=16,
    ) -> "ImageStack":
        """Generate a synthetic ImageStack from a set of Features stored in an IntensityTable

        Parameters
        ----------
        intensities : IntensityTable
            IntensityTable containing coordinates of fluorophores. Used to position and generate
            spots in the output ImageStack
        num_z : int
            Number of z-planes in the ImageStack
        height : int
            Height in pixels of the ImageStack
        width : int
            Width in pixels of the ImageStack
        n_photons_background : int
            Poisson rate for the number of background photons to add to each pixel of the image.
            Set this parameter to 0 to eliminate background.
            (default 1000)
        point_spread_function : Tuple[int]
            The width of the gaussian density wherein photons spread around their light source.
            Set to zero to eliminate this (default (4, 2, 2))
        camera_detection_efficiency : float
            The efficiency of the camera to detect light. Set to 1 to remove this filter (default
            0.25)
        background_electrons : int
            Poisson rate for the number of spurious electrons detected per pixel during image
            capture by the camera (default 1)
        graylevel : float
            The number of shades of gray displayable by the synthetic camera. Larger numbers will
            produce higher resolution images (default 37000 / 2 ** 16)
        ad_conversion_bits : int
            The number of bits used during analog to visual conversion (default 16)

        Returns
        -------
        ImageStack :
            synthetic spots

        """
        # check some params
        if not 0 < camera_detection_efficiency <= 1:
            raise ValueError(
                f'invalid camera_detection_efficiency value: {camera_detection_efficiency}. '
                f'Must be in the interval (0, 1].')

        def select_uint_dtype(array):
            """choose appropriate dtype based on values of an array"""
            max_val = np.max(array)
            for dtype in (np.uint8, np.uint16, np.uint32):
                if max_val <= np.iinfo(dtype).max:
                    return array.astype(dtype)
            raise ValueError('value exceeds dynamic range of largest skimage-supported type')

        # make sure requested dimensions are large enough to support intensity values
        indices = zip((Indices.Z.value, Indices.Y.value, Indices.X.value), (num_z, height, width))
        for index, requested_size in indices:
            required_size = intensities.coords[index].values.max()
            if required_size > requested_size:
                raise ValueError(
                    f'locations of intensities contained in table exceed the size of requested '
                    f'dimension {index}. Required size {required_size} > {requested_size}.')

        # create an empty array of the correct size
        image = np.zeros(
            (
                intensities.sizes[Indices.ROUND.value],
                intensities.sizes[Indices.CH.value],
                num_z,
                height,
                width
            ), dtype=np.uint32
        )

        # starfish uses float images, but the logic here requires uint. We cast, and will cast back
        # at the end of the function
        intensities.values = img_as_uint(intensities)

        for ch, round_ in product(*(range(s) for s in intensities.shape[1:])):
            spots = intensities[:, ch, round_]

            # numpy deprecated casting a specific way of casting floats that is triggered in xarray
            with warnings.catch_warnings():
                warnings.simplefilter('ignore', FutureWarning)
                values = spots.where(spots, drop=True)

            image[round_, ch, values.z, values.y, values.x] = values

        intensities.values = img_as_float32(intensities)

        # add imaging noise
        image += np.random.poisson(n_photons_background, size=image.shape).astype(np.uint32)

        # blur image over coordinates, but not over round_/channels (dim 0, 1)
        sigma = (0, 0) + point_spread_function
        image = gaussian_filter(image, sigma=sigma, mode='nearest')

        image = image * camera_detection_efficiency

        image += np.random.normal(scale=background_electrons, size=image.shape)

        # mimic analog to digital conversion
        image = (image / graylevel).astype(int).clip(0, 2 ** ad_conversion_bits)

        # clip in case we've picked up some negative values
        image = np.clip(image, 0, a_max=None)

        # set the smallest int datatype that supports the data's intensity range
        image = select_uint_dtype(image)

        # convert to float for ImageStack
        with warnings.catch_warnings():
            # possible precision loss when casting from uint to float is acceptable
            warnings.simplefilter('ignore', UserWarning)
            image = img_as_float32(image)

        return cls.from_numpy_array(image)



import copy

import numpy as np

from starfish.imagestack.imagestack import ImageStack
from starfish.types import Indices
from starfish.util.synthesize import SyntheticData


def divide(array, value):
    return array / value


def test_apply():
    """test that apply correctly applies a simple function across 2d tiles of a Stack"""
    stack = ImageStack.synthetic_stack()
    assert (stack.xarray == 1).all()
    output = stack.apply(divide, value=2)
    assert (output.xarray == 0.5).all()


def test_apply_3d():
    """test that apply correctly applies a simple function across 3d volumes of a Stack"""
    stack = ImageStack.synthetic_stack()
    assert np.all(stack.xarray == 1)
    stack.apply(divide, in_place=True, value=4,
                group_by={Indices.ROUND.value, Indices.CH.value})
    assert (stack.xarray == 0.25).all()


def test_apply_labeled_dataset():
    """
    test that apply correctly applies a simple function across starfish-generated synthetic data
    """
    original = SyntheticData().spots()
    image = original.apply(divide, value=2)
    assert np.all(image.xarray == original.xarray / 2)


def test_apply_in_place():
    """
    test that apply correctly applies a simple function across a starfish stack without modifying
    original data
    """
    image = SyntheticData().spots()
    original = copy.deepcopy(image)
    image.apply(divide, value=2, in_place=True)
    assert np.all(image.xarray == original.xarray / 2)



import os
import subprocess


def test_profiler():
    """Make sure that `starfish --profile works."""
    cmdline = [
        "starfish",
        "--profile",
        "version",
    ]
    env = os.environ.copy()
    subprocess.check_call(cmdline, env=env)



import numpy as np
import pytest
from scipy.ndimage.filters import gaussian_filter

from starfish.imagestack.imagestack import ImageStack
from starfish.spots._detector._base import SpotFinderAlgorithmBase
from starfish.spots._detector.blob import BlobDetector
from starfish.spots._detector.detect import detect_spots
from starfish.spots._detector.trackpy_local_max_peak_finder import TrackpyLocalMaxPeakFinder
from starfish.types import Indices


def simple_gaussian_spot_detector() -> BlobDetector:
    """create a basic gaussian spot detector"""
    return BlobDetector(min_sigma=1, max_sigma=4, num_sigma=5, threshold=0, measurement_type='max')


def simple_local_max_spot_detector() -> TrackpyLocalMaxPeakFinder:
    """create a basic local max peak finder"""
    return TrackpyLocalMaxPeakFinder(
        spot_diameter=3,
        min_mass=0.01,
        max_size=10,
        separation=2,
    )


# initialize spot detectors
gaussian_spot_detector = simple_gaussian_spot_detector()
local_max_spot_detector = simple_local_max_spot_detector()


def synthetic_two_spot_3d_2round_2ch() -> ImageStack:
    """produce a 2-channel 2-round ImageStack

    Notes
    -----
    - After Gaussian filtering, all max intensities are 7
    - Two spots are located at (4, 10, 90) and (6, 90, 10)
    - Both spots are 1-hot, and decode to:
        - spot 1: (round 0, ch 0), (round 1, ch 1)
        - spot 2: (round 0, ch 1), (round 1, ch 0)

    Returns
    -------
    ImageStack :
        noiseless ImageStack containing two spots

    """

    # blank data_image
    data = np.zeros((2, 2, 10, 100, 100), dtype=np.float32)

    # round 0 channel 0
    data[0, 0, 4, 10, 90] = 1.0
    data[0, 0, 5, 90, 10] = 0

    # round 0 channel 1
    data[0, 1, 4, 10, 90] = 0
    data[0, 1, 5, 90, 10] = 1.0

    # round 1 channel 0
    data[1, 0, 4, 10, 90] = 0
    data[1, 0, 5, 90, 10] = 1.0

    # round 1 channel 1
    data[1, 1, 4, 10, 90] = 1.0
    data[1, 1, 5, 90, 10] = 0

    data = gaussian_filter(data, sigma=(0, 0, 2, 2, 2))
    return ImageStack.from_numpy_array(data)


# create the data_stack
data_stack = synthetic_two_spot_3d_2round_2ch()


@pytest.mark.parametrize('data_stack, spot_detector, radius_is_gyration', [
    (data_stack, gaussian_spot_detector, False),
    (data_stack, local_max_spot_detector, True)
])
def test_spot_detection_with_reference_image(
        data_stack: ImageStack,
        spot_detector: SpotFinderAlgorithmBase,
        radius_is_gyration: bool,
):
    """
    This testing method uses a reference image to identify spot locations. Thus, it should detect
    two spots, each with max intensity 7. Because the channels and rounds are aggregated, this
    method should recognize the 1-hot code used in the testing data, and see one channel "on" per
    round. Thus, the total intensity across all channels and round for each spot should be 14.

    """
    reference_image_mp = data_stack.max_proj(Indices.CH, Indices.ROUND)
    reference_image_mp_numpy = reference_image_mp._squeezed_numpy(Indices.CH, Indices.ROUND)

    intensity_table = detect_spots(
        data_stack=data_stack,
        spot_finding_method=spot_detector.image_to_spots,
        reference_image=reference_image_mp_numpy,
        measurement_function=np.max,
        radius_is_gyration=radius_is_gyration,
    )
    assert intensity_table.shape == (2, 2, 2), "wrong number of spots detected"
    expected = [0.01587425, 0.01587425]
    assert np.allclose(intensity_table.sum((Indices.ROUND, Indices.CH)).values, expected), \
        "wrong spot intensities detected"


@pytest.mark.parametrize('data_stack, spot_detector, radius_is_gyration', [
    (data_stack, gaussian_spot_detector, False),
    (data_stack, local_max_spot_detector, True)
])
def test_spot_detection_with_reference_image_from_max_projection(
        data_stack: ImageStack,
        spot_detector: SpotFinderAlgorithmBase,
        radius_is_gyration: bool,
):
    """
    This testing method builds a reference image to identify spot locations. Thus, it should detect
    two spots, each with max intensity 7. Because the channels and rounds are aggregated, this
    method should recognize the 1-hot code used in the testing data, and see one channel "on" per
    round. Thus, the total intensity across all channels and round for each spot should be 14.
    """
    intensity_table = detect_spots(
        data_stack=data_stack,
        spot_finding_method=spot_detector.image_to_spots,
        reference_image_from_max_projection=True,
        measurement_function=np.max,
        radius_is_gyration=radius_is_gyration,
    )
    assert intensity_table.shape == (2, 2, 2), "wrong number of spots detected"
    expected = [0.01587425, 0.01587425]
    assert np.allclose(intensity_table.sum((Indices.ROUND, Indices.CH)).values, expected), \
        "wrong spot intensities detected"


@pytest.mark.parametrize('data_stack, spot_detector, radius_is_gyration', [
    (data_stack, gaussian_spot_detector, False),
    (data_stack, local_max_spot_detector, True)
])
def test_spot_finding_no_reference_image(
        data_stack: ImageStack,
        spot_detector: SpotFinderAlgorithmBase,
        radius_is_gyration: bool,
):
    """
    This testing method does not provide a reference image, and should therefore check for spots
    in each (round, ch) combination in sequence. With the given input, it should detect 4 spots,
    each with a max value of 7. Because each (round, ch) are measured sequentially, each spot only
    measures a single channel. Thus the total intensity across all rounds and channels for each
    spot should be 7.
    """
    intensity_table = detect_spots(
        data_stack=data_stack,
        spot_finding_method=spot_detector.image_to_spots,
        measurement_function=np.max,
        radius_is_gyration=radius_is_gyration,
    )
    assert intensity_table.shape == (4, 2, 2), "wrong number of spots detected"
    expected = [0.00793712] * 4
    assert np.allclose(intensity_table.sum((Indices.ROUND, Indices.CH)).values, expected), \
        "wrong spot intensities detected"



from copy import deepcopy
from typing import Optional, Tuple, Union

import click
import numpy as np
from scipy.ndimage import fourier_shift
from skimage.feature import register_translation

from starfish.image._filter.util import preserve_float_range
from starfish.imagestack.imagestack import ImageStack
from starfish.types import Indices
from ._base import RegistrationAlgorithmBase


class FourierShiftRegistration(RegistrationAlgorithmBase):
    """
    Implements fourier shift registration.  TODO: (dganguli) FILL IN DETAILS HERE PLS.

    Performs a simple translation registration.

    """
    def __init__(self, upsampling: int, reference_stack: Union[str, ImageStack], **kwargs) -> None:
        """Implements fourier shift registrations, which performs a simple translation registration

        Parameters
        ----------
        upsampling : int
            images are registered to within 1 / upsample_factor of a pixel
        reference_stack : ImageStack
            the ImageStack against which this object will register images

        See Also
        --------
        https://en.wikipedia.org/wiki/Phase_correlation

        """
        self.upsampling = upsampling

        # TODO ambrosejcarr: remove the ability to load from string in the constructor, move to CLI
        if isinstance(reference_stack, ImageStack):
            self.reference_stack = reference_stack
        else:
            self.reference_stack = ImageStack.from_path_or_url(reference_stack)

    def run(self, image: ImageStack, in_place: bool=False) -> Optional[ImageStack]:
        """Register an ImageStack against a reference image.

        Parameters
        ----------
        image : ImageStack
            The stack to be registered
        in_place : bool
            If false, return a new registered stack. Else, register in-place (default False)

        Returns
        -------


        """

        if not in_place:
            image = deepcopy(image)

        # TODO: (ambrosejcarr) is this the appropriate way of dealing with Z in registration?
        mp = image.max_proj(Indices.CH, Indices.Z)
        mp_numpy = mp._squeezed_numpy(Indices.CH, Indices.Z)
        reference_image_mp = self.reference_stack.max_proj(Indices.ROUND, Indices.CH, Indices.Z)
        reference_image_numpy = reference_image_mp._squeezed_numpy(Indices.ROUND,
                                                                   Indices.CH,
                                                                   Indices.Z)

        for r in range(image.num_rounds):
            # compute shift between maximum projection (across channels) and dots, for each round
            # TODO: make the max projection array ignorant of axes ordering.
            shift, error = compute_shift(mp_numpy[r, :, :], reference_image_numpy, self.upsampling)
            print(f"For round: {r}, Shift: {shift}, Error: {error}")

            for c in range(image.num_chs):
                for z in range(image.num_zlayers):
                    # apply shift to all zlayers, channels, and imaging rounds
                    indices = {Indices.ROUND: r, Indices.CH: c, Indices.Z: z}
                    data, axes = image.get_slice(indices=indices)
                    assert len(axes) == 0

                    result = shift_im(data, shift)
                    result = preserve_float_range(result)

                    image.set_slice(indices=indices, data=result)

        if not in_place:
            return image
        return None

    @staticmethod
    @click.command("FourierShiftRegistration")
    @click.option("--upsampling", default=1, type=int, help="Amount of up-sampling")
    @click.option("--reference-stack", required=True, type=click.Path(exists=True),
                  help="The image stack to align the input image stack to.")
    @click.pass_context
    def _cli(ctx, upsampling, reference_stack):
        ctx.obj["component"]._cli_run(ctx, FourierShiftRegistration(upsampling, reference_stack))


def compute_shift(
        im: np.ndarray, ref: np.ndarray, upsample_factor: int=1
) -> Tuple[np.ndarray, float]:
    """calculate subpixel image translation through cross-correlation

    Parameters
    ----------
    im : np.ndarray
        reference image
    ref : np.ndarray
        target image
    upsample_factor : int
        images are registered to within 1 / upsample_factor of a pixel

    Returns
    -------
    np.ndarray :
        shift vector required to register ref
    float :
        translation invariant normalized RMS error
    """
    shift, error, _ = register_translation(im, ref, upsample_factor)
    return shift, error


def shift_im(im: np.ndarray, shift: np.ndarray) -> np.ndarray:
    """register image according to the provided shift values"""
    fim_shift = fourier_shift(np.fft.fftn(im), shift * -1)
    im_shift = np.fft.ifftn(fim_shift)
    return im_shift.real



from functools import partial
from typing import Callable, Optional, Tuple, Union

import click
import numpy as np
import xarray as xr

from starfish.image._filter.gaussian_low_pass import GaussianLowPass
from starfish.imagestack.imagestack import ImageStack
from starfish.types import Number
from ._base import FilterAlgorithmBase
from .util import (
    determine_axes_to_group_by,
    preserve_float_range,
    validate_and_broadcast_kernel_size,
)


class GaussianHighPass(FilterAlgorithmBase):

    def __init__(
            self, sigma: Union[Number, Tuple[Number]], is_volume: bool=False,
    ) -> None:
        """Gaussian high pass filter

        Parameters
        ----------
        sigma : Union[Number, Tuple[Number]]
            standard deviation of gaussian kernel
        is_volume : bool
            If True, 3d (z, y, x) volumes will be filtered, otherwise, filter 2d tiles
            independently.

        """
        self.sigma = validate_and_broadcast_kernel_size(sigma, is_volume)
        self.is_volume = is_volume

    _DEFAULT_TESTING_PARAMETERS = {"sigma": 3}

    @staticmethod
    def _high_pass(
            image: Union[xr.DataArray, np.ndarray],
            sigma: Union[Number, Tuple[Number]],
            rescale: bool=False
    ) -> Union[xr.DataArray, np.ndarray]:
        """
        Applies a gaussian high pass filter to an image

        Parameters
        ----------
        image : numpy.ndarray[np.float32]
            2-d or 3-d image data
        sigma : Union[Number, Tuple[Number]]
            Standard deviation of gaussian kernel
        rescale : bool
            If true scales data by max value, if false clips max values to one

        Returns
        -------
        np.ndarray :
            filtered image of the same shape as the input image
        """

        blurred = GaussianLowPass._low_pass(image, sigma)
        filtered = image - blurred
        filtered = preserve_float_range(filtered, rescale)

        return filtered

    def run(
            self, stack: ImageStack, in_place: bool=False, verbose: bool=True,
            n_processes: Optional[int]=None
    ) -> ImageStack:
        """Perform filtering of an image stack

        Parameters
        ----------
        stack : ImageStack
            Stack to be filtered.
        in_place : bool
            if True, process ImageStack in-place, otherwise return a new stack
        verbose : bool
            if True, report on filtering progress (default = False)
        n_processes : Optional[int]
            Number of parallel processes to devote to calculating the filter

        Returns
        -------
        ImageStack :
            If in-place is False, return the results of filter as a new stack.  Otherwise return the
            original stack.

        """
        group_by = determine_axes_to_group_by(self.is_volume)
        high_pass: Callable = partial(self._high_pass, sigma=self.sigma)
        result = stack.apply(
            high_pass,
            group_by=group_by, verbose=verbose, in_place=in_place, n_processes=n_processes
        )
        return result

    @staticmethod
    @click.command("GaussianHighPass")
    @click.option("--sigma", type=float, help="standard deviation of gaussian kernel")
    @click.option("--is-volume", is_flag=True,
                  help="indicates that the image stack should be filtered in 3d")
    @click.pass_context
    def _cli(ctx, sigma, is_volume):
        ctx.obj["component"]._cli_run(ctx, GaussianHighPass(sigma, is_volume))



# Definition of the processing class
from functools import partial
from typing import Callable, Optional, Tuple, Union

import click
import numpy as np
from scipy.ndimage import gaussian_laplace

from starfish.image._filter._base import FilterAlgorithmBase
from starfish.image._filter.util import (
    determine_axes_to_group_by,
    preserve_float_range,
    validate_and_broadcast_kernel_size,
)
from starfish.imagestack.imagestack import ImageStack
from starfish.types import Number


class Laplace(FilterAlgorithmBase):
    """
    Multi-dimensional Laplace filter, using Gaussian second derivatives.
    This filter wraps scipy.ndimage.gaussian_laplace

    Parameters
    ----------
    sigma : Union[Number, Tuple[Number]]
        Standard deviation for Gaussian kernel.
    is_volume : bool
        If True, 3d (z, y, x) volumes will be filtered, otherwise, filter 2d tiles
        independently.

    """

    def __init__(
            self,
            sigma: Union[Number, Tuple[Number]], mode: str='reflect',
            cval: float=0.0, is_volume: bool=False,
    ) -> None:
        """Multi-dimensional gaussian-laplacian filter used to enhance dots against background

        Parameters
        ----------
        sigma_gauss : Union[Number, Tuple[Number]]
            Standard deviation for Gaussian kernel to enhance dots.

        mode: The mode parameter determines how the input array is extended when
            the filter overlaps a border. By passing a sequence of modes with
            length equal to the number of dimensions of the input array,
            different modes can be specified along each axis. Default value
            is ‘reflect’.
            The valid values and their behavior is as follows:

            ‘reflect’ (d c b a | a b c d | d c b a)
            The input is extended by reflecting about the edge of the last pixel.

            ‘constant’ (k k k k | a b c d | k k k k)
            The input is extended by filling all values beyond the edge with the same
            constant value, defined by the cval parameter.

            ‘nearest’ (a a a a | a b c d | d d d d)
            The input is extended by replicating the last pixel.

            ‘mirror’ (d c b | a b c d | c b a)
            The input is extended by reflecting about the center of the last pixel.

            ‘wrap’ (a b c d | a b c d | a b c d)
            The input is extended by wrapping around to the opposite edge.

        cval : scalar, optional
        Value to fill past edges of input if mode is ‘constant’. Default is 0.0.

        is_volume: bool
            True is the image is a stack
        """

        self.sigma = validate_and_broadcast_kernel_size(sigma, is_volume=is_volume)
        self.mode = mode
        self.cval = cval
        self.is_volume = is_volume

    _DEFAULT_TESTING_PARAMETERS = {"sigma": 0.5}

    @staticmethod
    def _gaussian_laplace(image: np.ndarray, sigma: Union[Number, Tuple[Number]],
                          mode: str = 'reflect', cval: float = 0.0) -> np.ndarray:
        filtered = gaussian_laplace(
            image, sigma=sigma, mode=mode, cval=cval)

        filtered = -filtered  # the peaks are negative so invert the signal
        filtered = preserve_float_range(filtered)

        return filtered

    def run(
            self, stack: ImageStack, in_place: bool = True, verbose: bool = True,
            n_processes: Optional[int]=None
    ) -> ImageStack:
        """Perform filtering of an image stack
        Parameters
        ----------
        stack : ImageStack
            Stack to be filtered.
        in_place : bool
            if True, process ImageStack in-place, otherwise return a new stack
        verbose : bool
            if True, report on filtering progress (default = False)
        Returns
        -------
        ImageStack :
            if in-place is False, return the results of filter as a new stack
        """
        group_by = determine_axes_to_group_by(self.is_volume)
        apply_filtering: Callable = partial(self._gaussian_laplace, sigma=self.sigma)
        return stack.apply(
            apply_filtering,
            group_by=group_by, verbose=verbose, in_place=in_place, n_processes=n_processes,
        )

    @staticmethod
    @click.command("Laplace")
    @click.option(
        "--sigma", type=float,
        help="Standard deviation of gaussian kernel for spot enhancement")
    @click.option(
        "--mode", default="reflect",
        help="How the input array is extended when the filter overlaps a border")
    @click.option(
        "--cval", default=0.0,
        help="Value to fill past edges of input if mode is ‘constant")
    @click.option(
        "--is-volume", is_flag=True,
        help="indicates that the image stack should be filtered in 3d")
    @click.pass_context
    def _cli(ctx, sigma, mode, cval, is_volume):
        ctx.obj["component"]._cli_run(ctx, Laplace(sigma, mode, cval, is_volume))



from starfish.pipeline.algorithmbase import AlgorithmBase


class DecoderAlgorithmBase(AlgorithmBase):
    def run(self, encoded, codebook):
        """Performs decoding on the spots found, using the codebook specified."""
        raise NotImplementedError()



from typing import Any, Dict, Optional

import matplotlib.axes
import matplotlib.figure
import matplotlib.pyplot as plt


# TODO ambrosejcarr: need to scale font size with figure size
# TODO ambrosejcarr: need a function to equalize plot axis sizes

def set_default_arguments(dictionary: Optional[Dict]=None, **kwargs) -> Dict[str, Any]:
    """
    Non-overwriting kwargs setter. Sets arguments in an optional dictionary only if they are not
    already present

    Parameters
    ----------
    dictionary : Dict
        dictionary to set keyword arguments in. Created if not provided
    kwargs : Dict
        keywords to add to the dictionary

    Returns
    -------
    Dict :
        Dictionary containing keywords

    """
    dictionary = {} if dictionary is None else dictionary
    for key, value in kwargs.items():
        if key not in dictionary:
            dictionary[key] = value
    return dictionary


def dpi_correction(figure: Optional[matplotlib.figure.Figure]=None) -> float:
    """
    matplotlib.scatter size correction for DPI to ensure that points have constant size across
    visible scales

    Parameters
    ----------
    figure : Optional[matplotlib.figure.Figure]
        Figure in which to scale scatterplot size. If not provided, defaults to current figure.

    Returns
    -------
    float :
        multiplier for matplotlib.plot.scatter markersize kwarg to ensure correct size in pixels

    """
    if figure is None:
        figure = plt.gcf()
    # 72 is the default pixel:point ratio for matplotlib
    point_multiplier = 72. / figure.dpi
    return point_multiplier


def remove_axis_frame(axis: Optional[matplotlib.axes.Axes]=None) -> matplotlib.axes.Axes:
    """Remove the spines and ticks from an axis.

    Parameters
    ----------
    axis : Optional[matplotlib.axes.Axes]
        Optional matplotlib axis. Defaults to current axis if not provided

    Returns
    -------

    """
    axis = axis if axis is not None else plt.gca()
    axis.set_axis_off()
    return axis


def annotate_axis(
        axis: Optional[matplotlib.axes.Axes]=None,
        title: Optional[str]=None,
        xlabel: Optional[str]=None,
        ylabel: Optional[str]=None,
        **kwargs
) -> matplotlib.axes.Axes:
    """Convenience parameter to set labels on an axis

    Parameters
    ----------
    axis : Optional[matplotlib.axes.Axes]
        Optional axis. Defaults to current Axis.
    title : Optional[str]
        title to set
    xlabel : Optional[str]
        xlabel to set
    ylabel : Optional[str]
        ylabel to set
    kwargs : Dict
        keyword arguments to pass to ALL annotations

    Returns
    -------

    """
    axis = axis if axis is not None else plt.gca()
    if title is not None:
        axis.set_title(title, **kwargs)
    if xlabel is not None:
        axis.set_xlabel(xlabel, **kwargs)
    if ylabel is not None:
        axis.set_ylabel(ylabel, **kwargs)
    return axis


def equalize_axes_extents(axis: Optional[matplotlib.axes.Axes]=None) -> matplotlib.axes.Axes:
    """set the x and y axes to the greater extent of both axes"""
    axis = axis if axis is not None else plt.gca()
    x_min, x_max = axis.get_ylim()
    y_min, y_max = axis.get_xlim()
    max_ = max(x_max, y_max)
    min_ = min(x_min, x_max)
    axis.set_xlim(min_, max_)
    axis.set_ylim(min_, max_)
    return axis



"""
This module implements default providers of data to the experiment builders.
"""

from typing import Mapping, Tuple, Type, Union

import numpy as np
from slicedimage import (
    ImageFormat,
)

from starfish.types import Coordinates, Number
from .providers import FetchedTile, TileFetcher


class RandomNoiseTile(FetchedTile):
    """
    This is a simple implementation of :class:`.FetchedImage` that simply regenerates random data
    for the image.
    """
    @property
    def shape(self) -> Tuple[int, ...]:
        return 1536, 1024

    @property
    def coordinates(self) -> Mapping[Union[str, Coordinates], Union[Number, Tuple[Number, Number]]]:
        return {
            Coordinates.X: (0.0, 0.0001),
            Coordinates.Y: (0.0, 0.0001),
            Coordinates.Z: (0.0, 0.0001),
        }

    @property
    def format(self) -> ImageFormat:
        return ImageFormat.TIFF

    def tile_data(self) -> np.ndarray:
        return np.random.randint(0, 256, size=self.shape, dtype=np.uint8)


class OnesTile(FetchedTile):
    """
    This is a simple implementation of :class:`.FetchedImage` that simply is entirely all pixels at
    maximum intensity.
    """
    def __init__(self, shape: Tuple[int, int]) -> None:
        super().__init__()
        self._shape = shape

    @property
    def shape(self) -> Tuple[int, ...]:
        return self._shape

    @property
    def coordinates(self) -> Mapping[Union[str, Coordinates], Union[Number, Tuple[Number, Number]]]:
        return {
            Coordinates.X: (0.0, 0.0001),
            Coordinates.Y: (0.0, 0.0001),
            Coordinates.Z: (0.0, 0.0001),
        }

    @property
    def format(self) -> ImageFormat:
        return ImageFormat.TIFF

    def tile_data(self) -> np.ndarray:
        return np.full(shape=self.shape, fill_value=1.0, dtype=np.float32)


def tile_fetcher_factory(
        fetched_tile_cls: Type[FetchedTile],
        pass_tile_indices: bool=False,
        *fetched_tile_constructor_args,
        **fetched_tile_constructor_kwargs,
) -> TileFetcher:
    """
    Given a class of that implements :class:`.FetchedTile`, return a TileFetcher that returns an
    instance of that class.  If `pass_tile_indices` is True, then the TileFetcher is constructed
    with fov/round/ch/z.  The constructor is always invoked with variable-length arguments from
    `fetched_tile_constructor_args` and keyword arguments from `fetched_tile_constructor_kwargs`.
    """
    class ResultingClass(TileFetcher):
        def get_tile(self, fov: int, r: int, ch: int, z: int) -> FetchedTile:
            args = list()
            if pass_tile_indices:
                args.extend([fov, r, ch, z])
            args.extend(fetched_tile_constructor_args)

            return fetched_tile_cls(*args, **fetched_tile_constructor_kwargs)

    return ResultingClass()



"""Constants to support formatting for serialization and deserialization of codebooks."""

from semantic_version import Version

CURRENT_VERSION = Version("0.0.0")
MIN_SUPPORTED_VERSION = Version("0.0.0")
MAX_SUPPORTED_VERSION = Version("0.0.0")


class DocumentKeys:
    VERSION_KEY = 'version'
    MAPPINGS_KEY = 'mappings'



"""
Notes
-----
This test and docs/source/usage/iss/iss_cli.sh test the same code paths and should be updated
together
"""
import os
import sys
import unittest

import numpy as np
import pandas as pd

from starfish.test.full_pipelines.cli._base_cli_test import CLITest
from starfish.types import Features


class TestWithIssData(CLITest, unittest.TestCase):

    @property
    def spots_file(self):
        return "decoded-spots.nc"

    @property
    def subdirs(self):
        return (
            "raw",
            "formatted",
            "registered",
            "filtered",
            "results",
        )

    @property
    def stages(self):
        return (
            [
                sys.executable,
                "examples/get_cli_test_data.py",
                "https://d2nhj9g34unfro.cloudfront.net/20181005/ISS-TEST/",
                lambda tempdir, *args, **kwargs: os.path.join(tempdir, "formatted")
            ],
            [
                "starfish", "registration",
                "--input", lambda tempdir, *args, **kwargs: os.path.join(
                    tempdir, "formatted/fov_001", "hybridization.json"),
                "--output", lambda tempdir, *args, **kwargs: os.path.join(
                    tempdir, "registered", "hybridization.json"),
                "FourierShiftRegistration",
                "--reference-stack", lambda tempdir, *args, **kwargs: os.path.join(
                    tempdir, "formatted/fov_001", "dots.json"),
                "--upsampling", "1000",
            ],
            [
                "starfish", "filter",
                "--input", lambda tempdir, *args, **kwargs: os.path.join(
                    tempdir, "registered", "hybridization.json"),
                "--output", lambda tempdir, *args, **kwargs: os.path.join(
                    tempdir, "filtered", "hybridization.json"),
                "WhiteTophat",
                "--masking-radius", "15",
            ],
            [
                "starfish", "filter",
                "--input", lambda tempdir, *args, **kwargs: os.path.join(
                    tempdir, "formatted/fov_001", "nuclei.json"),
                "--output", lambda tempdir, *args, **kwargs: os.path.join(
                    tempdir, "filtered", "nuclei.json"),
                "WhiteTophat",
                "--masking-radius", "15",
            ],
            [
                "starfish", "filter",
                "--input", lambda tempdir, *args, **kwargs: os.path.join(
                    tempdir, "formatted/fov_001", "dots.json"),
                "--output", lambda tempdir, *args, **kwargs: os.path.join(
                    tempdir, "filtered", "dots.json"),
                "WhiteTophat",
                "--masking-radius", "15",
            ],
            [
                "starfish", "detect_spots",
                "--input", lambda tempdir, *args, **kwargs: os.path.join(
                    tempdir, "filtered", "hybridization.json"),
                "--output", lambda tempdir, *args, **kwargs: os.path.join(
                    tempdir, "results", "spots.nc"),
                "--blobs-stack", lambda tempdir, *args, **kwargs: os.path.join(
                    tempdir, "filtered", "dots.json"),
                "BlobDetector",
                "--min-sigma", "4",
                "--max-sigma", "6",
                "--num-sigma", "20",
                "--threshold", "0.01",
            ],
            [
                "starfish", "segment",
                "--primary-images", lambda tempdir, *args, **kwargs: os.path.join(
                    tempdir, "filtered", "hybridization.json"),
                "--nuclei", lambda tempdir, *args, **kwargs: os.path.join(
                    tempdir, "filtered", "nuclei.json"),
                "-o", lambda tempdir, *args, **kwargs: os.path.join(
                    tempdir, "results", "label_image.png"),
                "Watershed",
                "--nuclei-threshold", ".16",
                "--input-threshold", ".22",
                "--min-distance", "57",
            ],
            [
                "starfish", "target_assignment",
                "--label-image",
                lambda tempdir, *args, **kwargs: os.path.join(
                    tempdir, "results", "label_image.png"),
                "--intensities", lambda tempdir, *args, **kwargs: os.path.join(
                    tempdir, "results", "spots.nc"),
                "--output", lambda tempdir, *args, **kwargs: os.path.join(
                    tempdir, "results", "targeted-spots.nc"),
                "Label",
            ],
            [
                "starfish", "decode",
                "-i", lambda tempdir, *args, **kwargs: os.path.join(
                    tempdir, "results", "targeted-spots.nc"),
                "--codebook", lambda tempdir, *args, **kwargs: os.path.join(
                    tempdir, "formatted", "codebook.json"),
                "-o", lambda tempdir, *args, **kwargs: os.path.join(
                    tempdir, "results", "decoded-spots.nc"),
                "PerRoundMaxChannelDecoder",
            ],
        )

    def verify_results(self, intensities):
        # TODO make this test stronger
        genes, counts = np.unique(
            intensities.coords[Features.TARGET], return_counts=True)
        gene_counts = pd.Series(counts, genes)
        # TODO THERE"S NO HUMAN/MOUSE KEYS?
        assert gene_counts['ACTB']



from functools import partial
from itertools import product
from typing import Callable, Dict, Sequence, Tuple, Union

import numpy as np
import pandas as pd
import xarray as xr

from starfish.imagestack.imagestack import ImageStack
from starfish.intensity_table.intensity_table import IntensityTable
from starfish.types import Features, Indices, Number, SpotAttributes


def measure_spot_intensity(
        image: Union[np.ndarray, xr.DataArray],
        spots: SpotAttributes,
        measurement_function: Callable[[Sequence], Number],
        radius_is_gyration: bool=False,
) -> pd.Series:
    """measure the intensity of each spot in spots in the corresponding image

    Parameters
    ----------
    image : Union[np.ndarray, xr.DataArray],
        3-d volume in which to measure intensities
    spots : pd.DataFrame
        SpotAttributes table containing coordinates and radii of spots
    measurement_function : Callable[[Sequence], Number])
        Function to apply over the spot volumes to identify the intensity (e.g. max, mean, ...)
    radius_is_gyration : bool
        if True, indicates that the radius corresponds to radius of gyration, which is a function of
        spot intensity, but typically is a smaller unit than the sigma generated by blob_log.
        In this case, the spot's bounding box is rounded up instead of down when measuring
        intensity. (default False)

    Returns
    -------
    pd.Series :
        Intensities for each spot in SpotAttributes

    """

    def fn(row: pd.Series) -> Number:
        data = image[
            row['z_min']:row['z_max'],
            row['y_min']:row['y_max'],
            row['x_min']:row['x_max']
        ]
        return measurement_function(data)

    if radius_is_gyration:
        radius = np.ceil(spots.data[Features.SPOT_RADIUS]).astype(int) + 1  # round up
    else:
        radius = spots.data[Features.SPOT_RADIUS].astype(int)  # truncate down to nearest integer
    for v, max_size in zip(['z', 'y', 'x'], image.shape):
        # numpy does exclusive max indexing, so need to subtract 1 from min to get centered box
        spots.data[f'{v}_min'] = np.clip(spots.data[v] - (radius - 1), 0, None)
        spots.data[f'{v}_max'] = np.clip(spots.data[v] + radius, None, max_size)
    return spots.data[['z_min', 'z_max', 'y_min', 'y_max', 'x_min', 'x_max']].astype(int).apply(
        fn,
        axis=1
    )


def measure_spot_intensities(
        data_image: ImageStack,
        spot_attributes: SpotAttributes,
        measurement_function: Callable[[Sequence], Number],
        radius_is_gyration: bool=False,
) -> IntensityTable:
    """given spots found from a reference image, find those spots across a data_image

    Parameters
    ----------
    data_image : ImageStack
        ImageStack containing multiple volumes for which spots' intensities must be calculated
    spot_attributes : pd.Dataframe
        Locations and radii of spots
    measurement_function : Callable[[Sequence], Number])
        Function to apply over the spot volumes to identify the intensity (e.g. max, mean, ...)
    radius_is_gyration : bool
        if True, indicates that the radius corresponds to radius of gyration, which is a function of
        spot intensity, but typically is a smaller unit than the sigma generated by blob_log.
        In this case, the spot's bounding box is rounded up instead of down when measuring
        intensity. (default False)

    Returns
    -------
    IntensityTable :
        3d tensor of (spot, channel, round) information for each coded spot

    """

    # determine the shape of the intensity table
    n_ch = data_image.shape[Indices.CH]
    n_round = data_image.shape[Indices.ROUND]

    # construct the empty intensity table
    intensity_table = IntensityTable.empty_intensity_table(
        spot_attributes=spot_attributes,
        n_ch=n_ch,
        n_round=n_round,
    )

    # fill the intensity table
    indices = product(range(n_ch), range(n_round))
    for c, r in indices:
        image, _ = data_image.get_slice({Indices.CH: c, Indices.ROUND: r})
        blob_intensities: pd.Series = measure_spot_intensity(
            image,
            spot_attributes,
            measurement_function,
            radius_is_gyration=radius_is_gyration
        )
        intensity_table[:, c, r] = blob_intensities

    return intensity_table


def concatenate_spot_attributes_to_intensities(
        spot_attributes: Sequence[Tuple[SpotAttributes, Dict[Indices, int]]]
) -> IntensityTable:
    """
    Merge multiple spot attributes frames into a single IntensityTable without merging across
    channels and imaging rounds

    Parameters
    ----------
    spot_attributes : Sequence[Tuple[SpotAttributes, Dict[Indices, int]]]
        A sequence of SpotAttribute objects and the Indices (channel, round) that each object is
        associated with.

    Returns
    -------
    IntensityTable :
        concatenated input SpotAttributes, converted to an IntensityTable object

    """
    n_ch: int = max(inds[Indices.CH] for _, inds in spot_attributes) + 1
    n_round: int = max(inds[Indices.ROUND] for _, inds in spot_attributes) + 1

    all_spots = pd.concat([sa.data for sa, inds in spot_attributes])
    # this drop call ensures only x, y, z, radius, and quality, are passed to the IntensityTable
    features_coordinates = all_spots.drop(['spot_id', 'intensity'], axis=1)

    intensity_table = IntensityTable.empty_intensity_table(
        SpotAttributes(features_coordinates), n_ch, n_round,
    )

    i = 0
    for attrs, inds in spot_attributes:
        for _, row in attrs.data.iterrows():
            intensity_table[i, inds[Indices.CH], inds[Indices.ROUND]] = row['intensity']
            i += 1

    return intensity_table


def detect_spots(
        data_stack: ImageStack,
        spot_finding_method: Callable[..., SpotAttributes],
        spot_finding_kwargs: Dict=None,
        reference_image: Union[xr.DataArray, np.ndarray]=None,
        reference_image_from_max_projection: bool=False,
        measurement_function: Callable[[Sequence], Number]=np.max,
        radius_is_gyration: bool=False,
        is_volume: bool=True
) -> IntensityTable:
    """Apply a spot_finding_method to a ImageStack

    Parameters
    ----------
    data_stack : ImageStack
        The ImageStack containing spots
    spot_finding_method : Callable[..., IntensityTable]
        The method to identify spots
    spot_finding_kwargs : Dict
        additional keyword arguments to pass to spot_finding_method
    reference_image : xr.DataArray
        (Optional) a reference image. If provided, spots will be found in this image, and then
        the locations that correspond to these spots will be measured across each channel and round,
        filling in the values in the IntensityTable
    reference_image_from_max_projection : Tuple[Indices]
        (Optional) if True, create a reference image by max-projecting the channels and imaging
        rounds found in data_image.
    measurement_function : Callable[[Sequence], Number]
        the function to apply over the spot area to extract the intensity value (default 'np.max')
    radius_is_gyration : bool
        if True, indicates that the radius corresponds to radius of gyration, which is a function of
        spot intensity, but typically is a smaller unit than the sigma generated by blob_log.
        In this case, the spot's bounding box is rounded up instead of down when measuring
        intensity. (default False)
    is_volume: bool
        If True, pass 3d volumes (x, y, z) to func, else pass 2d tiles (x, y) to func. (default
        True)

    Notes
    -----
    - This class will always detect spots in 3d. If 2d spot detection is desired, the data should
      be projected down to "fake 3d" prior to submission to this function
    - If neither reference_image nor reference_from_max_projection are passed, spots will be
      detected _independently_ in each channel. This assumes a non-multiplex imaging experiment,
      as only one (ch, round) will be measured for each spot.

    Returns
    -------
    IntensityTable :
        IntensityTable containing the intensity of each spot, its radius, and location in pixel
        coordinates

    """

    if spot_finding_kwargs is None:
        spot_finding_kwargs = {}

    if reference_image is not None and reference_image_from_max_projection:
        raise ValueError(
            'Please pass only one of reference_image and reference_image_from_max_projection'
        )

    if reference_image_from_max_projection:
        reference_image = data_stack.max_proj(Indices.CH, Indices.ROUND)
        reference_image = reference_image._squeezed_numpy(Indices.CH, Indices.ROUND)

    if is_volume:
        group_by = {Indices.ROUND, Indices.CH}
    else:
        group_by = {Indices.ROUND, Indices.CH, Indices.Z}

    if reference_image is not None:
        reference_spot_locations = spot_finding_method(reference_image, **spot_finding_kwargs)
        intensity_table = measure_spot_intensities(
            data_image=data_stack,
            spot_attributes=reference_spot_locations,
            measurement_function=measurement_function,
            radius_is_gyration=radius_is_gyration,
        )
    else:  # don't use a reference image, measure each
        spot_finding_method = partial(spot_finding_method, **spot_finding_kwargs)
        spot_attributes_list = data_stack.transform(
            func=spot_finding_method,
            group_by=group_by
        )
        intensity_table = concatenate_spot_attributes_to_intensities(spot_attributes_list)

    return intensity_table



from typing import Tuple, Union

import numpy as np
import pytest

from starfish.image._filter import gaussian_high_pass, mean_high_pass
from starfish.imagestack.imagestack import ImageStack
from starfish.types import Number


def random_data_image_stack_factory():
    data = np.random.uniform(0, 1, 100).reshape(1, 1, 1, 10, 10).astype(np.float32)
    return ImageStack.from_numpy_array(data)


@pytest.mark.parametrize('sigma, is_volume', [
    (1, False),
    ((1, 1), False),
    ((1, 0, 1), True)
])
def test_gaussian_high_pass(sigma: Union[Number, Tuple[Number]], is_volume: bool) -> None:
    """high pass is subtractive, sum of array should be less after running."""
    image_stack = random_data_image_stack_factory()
    sum_before = np.sum(image_stack.xarray)
    ghp = gaussian_high_pass.GaussianHighPass(sigma=sigma, is_volume=is_volume)
    result = ghp.run(image_stack)
    assert np.sum(result.xarray) < sum_before

@pytest.mark.parametrize('size, is_volume', [
    (1, False),
    ((1, 1), False),
    ((1, 0, 1), True)
])
def test_mean_high_pass(size: Union[Number, Tuple[Number]], is_volume: bool) -> None:
    """high pass is subtractive, sum of array should be less after running."""
    image_stack = random_data_image_stack_factory()
    sum_before = np.sum(image_stack.xarray)
    mhp = mean_high_pass.MeanHighPass(size=size, is_volume=is_volume)
    result = mhp.run(image_stack)
    assert np.sum(result.xarray) < sum_before



from starfish.imagestack.imagestack import ImageStack
from starfish.pipeline.algorithmbase import AlgorithmBase


class FilterAlgorithmBase(AlgorithmBase):
    def run(self, stack: ImageStack) -> ImageStack:
        """Performs filtering on an ImageStack."""
        raise NotImplementedError()



import json
import os
from typing import (
    Callable,
    Dict,
    MutableMapping,
    MutableSequence,
    MutableSet,
    Optional,
    Sequence,
    Set,
    Union,
)

from semantic_version import Version
from slicedimage import Collection, TileSet
from slicedimage.io import Reader, resolve_path_or_url, resolve_url
from slicedimage.urlpath import pathjoin

from sptx_format import validate_sptx
from starfish.codebook.codebook import Codebook
from starfish.imagestack.imagestack import ImageStack
from starfish.util.config import Config
from .version import MAX_SUPPORTED_VERSION, MIN_SUPPORTED_VERSION


class FieldOfView:
    """
    This encapsulates a field of view.  It contains the primary image and auxiliary images that are
    associated with the field of view.

    All images can be accessed using a key, i.e., FOV[aux_image_type].  The primary image is
    accessed using the key :py:attr:`starfish.experiment.experiment.FieldOFView.PRIMARY_IMAGES`.

    Attributes
    ----------
    name : str
        The name of the FOV.
    image_types : Set[str]
        A set of all the image types.
    """

    PRIMARY_IMAGES = 'primary'

    def __init__(
            self,
            name: str,
            images: Optional[MutableMapping[str, ImageStack]]=None,
            image_tilesets: Optional[MutableMapping[str, TileSet]]=None,
    ) -> None:
        """
        Fields of views can obtain their primary image from either an ImageStack or a TileSet (but
        only one).  It can obtain their auxiliary image dictionary from either a dictionary of
        auxiliary image name to ImageStack or a dictionary of auxiliary image name to TileSet (but
        only one).

        Note that if the source image is from a TileSet, the decoding of TileSet to ImageStack does
        not happen until the image is accessed.  Be prepared to handle errors when images are
        accessed.
        """
        self._name = name
        self._images: MutableMapping[str, Union[ImageStack, TileSet]]
        if images is not None:
            self._images = images
            if image_tilesets is not None:
                raise ValueError(
                    "Only one of (images, image_tilesets) should be set.")
        elif image_tilesets is not None:
            self._images = image_tilesets
        else:
            self._images = dict()

    def __repr__(self):
        images = '\n    '.join(
            f'{k}: {v}'
            for k, v in self._images.items()
            if k != FieldOfView.PRIMARY_IMAGES
        )
        return (
            f"<starfish.FieldOfView>\n"
            f"  Primary Image: {self._images[FieldOfView.PRIMARY_IMAGES]}\n"
            f"  Auxiliary Images:\n"
            f"    {images}"
        )

    @property
    def name(self) -> str:
        return self._name

    @property
    def image_types(self) -> Set[str]:
        return set(self._images.keys())

    def __getitem__(self, item) -> ImageStack:
        if isinstance(self._images[item], TileSet):
            self._images[item] = ImageStack(self._images[item])
        return self._images[item]


class Experiment:
    """
    This encapsulates an experiment, with one or more fields of view and a codebook.  An individual
    FOV can be retrieved using a key, i.e., experiment[fov_name].

    Methods
    -------
    from_json()
        Given a URL or a path to an experiment.json document, return an Experiment object
        corresponding to the document.
    fov()
        Given a callable that accepts a FOV, return the first FOVs that the callable returns True
        when passed the FOV.  Because there is no guaranteed sorting for the FOVs, use this
        cautiously.
    fovs()
        Given a callable that accepts a FOV, return all the FOVs that the callable returns True when
        passed the FOV.
    fovs_by_name()
        Given one or more FOV names, return the FOVs that match those names.

    Attributes
    ----------
    codebook : Codebook
        Returns the codebook associated with this experiment.
    extras : Dict
        Returns the extras dictionary associated with this experiment.
    """
    def __init__(
            self,
            fovs: Sequence[FieldOfView],
            codebook: Codebook,
            extras: dict,
            *,
            src_doc: dict=None,
    ) -> None:
        self._fovs = fovs
        self._codebook = codebook
        self._extras = extras
        self._src_doc = src_doc

    def __repr__(self):

        # truncate the list of fields of view if it is longer than print_n_fov
        print_n_fov = 4
        n_fields_of_view = list(self.items())[:print_n_fov]
        fields_of_view_str = "\n".join(
            f'{k}: {v}' for k, v in n_fields_of_view
        )

        # add an ellipsis if not all fields of view are being printed
        if len(self._fovs) > print_n_fov:
            fov_repr = f"{{\n{fields_of_view_str}\n  ...,\n}}"
        else:
            fov_repr = f"{{\n{fields_of_view_str}\n}}"

        # return the formatted string
        object_repr = f"<starfish.Experiment (FOVs={len(self._fovs)})>\n"
        return object_repr + fov_repr

    @classmethod
    def from_json(cls,
                  json_url: str,
                  strict: bool=None,
                  config: Optional[Union[str, Dict]]=None) -> "Experiment":
        """
        Construct an `Experiment` from an experiment.json file format specifier

        Parameters
        ----------
        json_url : str
            file path or web link to an experiment.json file
        strict : bool
            if true, then all JSON loaded by this method will be
            passed to the appropriate validator
        config : str or dict
            configuration property that will be passed to
            starfish.util.config.Config
        STARISH_CONFIG :
            This parameter is read from the environment to permit setting configuration
            values either directly or via a file. Keys read include:
             - cache.allow_caching
        STARFISH_STRICT_LOADING :
             This parameter is read from the environment. If set, then all JSON loaded by this
             method will be passed to the appropriate validator. The `strict` parameter to this
             method has priority over the environment variable.

        Returns
        -------
        Experiment :
            Experiment object serving the requested experiment data

        """
        if strict is None:
            strict = "STARFISH_STRICT_LOADING" in os.environ
        if strict:
            valid = validate_sptx.validate(json_url)
            if not valid:
                raise Exception("validation failed")

        config_obj = Config(config)  # STARFISH_CONFIG is assumed
        allow_caching = config_obj.lookup(["cache", "allow_caching"], True)

        backend, name, baseurl = resolve_path_or_url(json_url, allow_caching)
        with backend.read_contextmanager(name) as fh:
            experiment_document = json.load(fh)

        version = cls.verify_version(experiment_document['version'])

        _, codebook_name, codebook_baseurl = resolve_url(experiment_document['codebook'],
                                                         baseurl, allow_caching)
        codebook_absolute_url = pathjoin(codebook_baseurl, codebook_name)
        codebook = Codebook.from_json(codebook_absolute_url)

        extras = experiment_document['extras']

        fovs: MutableSequence[FieldOfView] = list()
        fov_tilesets: MutableMapping[str, TileSet]
        if version < Version("5.0.0"):
            primary_image: Collection = Reader.parse_doc(experiment_document['primary_images'],
                                                         baseurl)
            auxiliary_images: MutableMapping[str, Collection] = dict()
            for aux_image_type, aux_image_url in experiment_document['auxiliary_images'].items():
                auxiliary_images[aux_image_type] = Reader.parse_doc(aux_image_url, baseurl)

            for fov_name, primary_tileset in primary_image.all_tilesets():
                fov_tilesets = dict()
                fov_tilesets[FieldOfView.PRIMARY_IMAGES] = primary_tileset
                for aux_image_type, aux_image_collection in auxiliary_images.items():
                    aux_image_tileset = aux_image_collection.find_tileset(fov_name)
                    if aux_image_tileset is not None:
                        fov_tilesets[aux_image_type] = aux_image_tileset

                fov = FieldOfView(fov_name, image_tilesets=fov_tilesets)
                fovs.append(fov)
        else:
            images: MutableMapping[str, Collection] = dict()
            all_fov_names: MutableSet[str] = set()
            for image_type, image_url in experiment_document['images'].items():
                image = Reader.parse_doc(image_url, baseurl)
                images[image_type] = image
                for fov_name, _ in image.all_tilesets():
                    all_fov_names.add(fov_name)

            for fov_name in all_fov_names:
                fov_tilesets = dict()
                for image_type, image_collection in images.items():
                    image_tileset = image_collection.find_tileset(fov_name)
                    if image_tileset is not None:
                        fov_tilesets[image_type] = image_tileset

                fov = FieldOfView(fov_name, image_tilesets=fov_tilesets)
                fovs.append(fov)

        return Experiment(fovs, codebook, extras, src_doc=experiment_document)

    @classmethod
    def verify_version(cls, semantic_version_str: str) -> Version:
        version = Version(semantic_version_str)
        if not (MIN_SUPPORTED_VERSION <= version <= MAX_SUPPORTED_VERSION):
            raise ValueError(
                f"version {version} not supported.  This version of the starfish library only "
                f"supports formats from {MIN_SUPPORTED_VERSION} to "
                f"{MAX_SUPPORTED_VERSION}")
        return version

    def fov(
            self,
            filter_fn: Callable[[FieldOfView], bool]=lambda _: True,
    ) -> FieldOfView:
        """
        Given a callable filter_fn, apply it to all the FOVs in this experiment.  Return the first
        FOV such that filter_fn(FOV) returns True.  Because there is no guaranteed order for the
        FOVs, use this cautiously.

        If no FOV matches, raise LookupError.
        """
        for fov in self._fovs:
            if filter_fn(fov):
                return fov
        raise LookupError("Cannot find any FOV that the filter allows.")

    def fovs(
            self,
            filter_fn: Callable[[FieldOfView], bool]=lambda _: True,
    ) -> Sequence[FieldOfView]:
        """
        Given a callable filter_fn, apply it to all the FOVs in this experiment.  Return a list of
        FOVs such that filter_fn(FOV) returns True.
        """
        results: MutableSequence[FieldOfView] = list()
        for fov in self._fovs:
            if not filter_fn(fov):
                continue

            results.append(fov)
        return results

    def fovs_by_name(self, *names):
        """
        Given a callable filter_fn, apply it to all the FOVs in this experiment.  Return a list of
        FOVs such that filter_fn(FOV) returns True.
        """
        return self.fovs(filter_fn=lambda fov: fov.name in names)

    def __getitem__(self, item):
        fovs = self.fovs_by_name(item)
        if len(fovs) == 0:
            raise IndexError(f"No field of view with name \"{item}\"")
        return fovs[0]

    def keys(self):
        return (fov.name for fov in self.fovs())

    def values(self):
        return (fov for fov in self.fovs())

    def items(self):
        return ((fov.name, fov) for fov in self.fovs())

    @property
    def codebook(self) -> Codebook:
        return self._codebook

    @property
    def extras(self):
        return self._extras
